;			The body of KANREN
;
; The appropriate prelude (e.g., chez-specific.scm) is assumed.
;
; $Id: binder-kanren.scm,v 1.1 2004/07/10 04:15:56 oleg Exp $

(define errorf error)
(define pretty-print pp)
(define (time x) x)

(define-syntax lambda@
  (syntax-rules ()
    [(_ (formal) body0 body1 ...) (lambda (formal) body0 body1 ...)]
    [(_ (formal0 formal1 formal2 ...) body0 body1 ...)
     (lambda (formal0)
       (lambda@ (formal1 formal2 ...) body0 body1 ...))]))

; (define-syntax trace-lambda@
;   (syntax-rules ()
;     [(_ id () body0 body1 ...) (begin body0 body1 ...)]
;     [(_ id (formal0 formal1 ...) body0 body1 ...)
;      (trace-lambda id (formal0)
;        (trace-lambda@ id (formal1 ...) body0 body1 ...))]))

(define-syntax @  
  (syntax-rules ()
    [(_ rator rand) (rator rand)]
    [(_ rator rand0 rand1 rand2 ...) (@ (rator rand0) rand1 rand2 ...)]))

; (define symbol-append
;   (lambda symbs
;     (string->symbol
;       (apply string-append
;         (map symbol->string symbs)))))

(define Y
  (lambda (f)
    ((lambda (u) (u (lambda (x) (lambda (n) ((f (u x)) n)))))
     (lambda (x) (x x)))))


; (define-syntax @    
;   (syntax-rules (syntax-rules)
;     [(_ (syntax-rules sdata ...) rand0 ...)
;       (let-syntax 
; 	((tempname (syntax-rules sdata ...)))
; 	(tempname rand0 ...))]
;     [(_ rator rand0 rand1 ...)
;      (@-simple rator rand0 rand1 ...)]))

; LET*-AND: a simplified and streamlined AND-LET*.
; The latter is defined in SRFI-2 <http://srfi.schemers.org/srfi-2/>

(define-syntax let*-and
  (syntax-rules ()
    [(_ false-exp () body0 body1 ...) (begin body0 body1 ...)]
    [(_ false-exp ([var0 exp0] [var1 exp1] ...) body0 body1 ...)
     (let ([var0 exp0])
       (if var0
         (let*-and false-exp ([var1 exp1] ...) body0 body1 ...)
         false-exp))]))

; Regression testing framework
; test-check TITLE TESTED-EXPRESSION EXPECTED-RESULT 
; where TITLE is something printable (e.g., a symbol or a string)
; EXPECTED-RESULT and TESTED-EXPRESSION are both expressions.
; The expressions are evaluated and their results are cmpared
; by equal?
; If the results compare, we just print the TITLE.
; Otherwise, we print the TITLE, the TESTED-EXPRESSION, and
; the both results.
(define-syntax test-check
  (syntax-rules ()
    [(_ title tested-expression expected-result)
     (begin
       (cout "Testing " title nl)
       (let* ([expected expected-result]
              [produced tested-expression])
         (or (equal? expected produced)
             (errorf 'test-check
               "Failed: ~a~%Expected: ~a~%Computed: ~a~%"
               'tested-expression expected produced))))]))

(test-check 'test-@-lambda@
  (@ (lambda@ (x y z) (+ x (+ y z))) 1 2 3)
  6)

'(test-check 'test-@-lambda@
  (@ (lambda@ (x y z) (+ x (+ y z))) 1 2 3)
  42)

;------------------------------------------------------------------------
; Logical variables, substitutions, and commitments (aka bindings)
   
(cond-expand
  (chez
    (define-record logical-variable (id) ()))
  (else
    ; use SRFI-9 records
    (define-record-type *logical-variable*
      (make-logical-variable name) logical-variable?
      (name logical-variable-id))
    ))
(define logical-variable make-logical-variable)
(define var? logical-variable?)


; Introduction of a logical variable
(define-syntax let-lv
  (syntax-rules ()
    [(_ () body) body]
    [(_ () body0 body1 body2 ...) (begin body0 body1 body2 ...)]
    [(_ (id ...) body0 body1 ...)
     (let ([id (logical-variable 'id)] ...) body0 body1 ...)]))


; (define logical-var-tag (list '*logical-var-tag*)) ; unique for eq?
; (define native-pair? pair?)
; (define logical-variable
;   (lambda (id)
;     (cons logical-var-tag id)))
; (define var?
;   (lambda (x)
;     (and (native-pair? x) (eq? (car x) logical-var-tag))))
; (define logical-variable-id
;   (lambda (x)
;     (if (var? x) (cdr x) 
;       (errorf 'logical-variable-id "Invalid Logic Variable: ~s" x))))
; (define pair?
;   (lambda (x)
;     (and (native-pair? x) (not (eq? (car x) logical-var-tag)))))


; Eigen-variables -- unique symbols that represent universally-quantified
; variables in a term
; For identification, we prefix the name of the eigen-variable with
; the exclamation mark. The mark makes sure the symbol stands out when
; printed.

(define eigen-variable
  (lambda (id)
    (symbol-append '! id '_ (gensym))))

(define eigen-var?
  (lambda (x)
    (and (symbol? x)
      (let ((str (symbol->string x)))
	(> (string-length str) 2)
	(char=? (string-ref str 0) #\!)))))


; (eigen (id ...) body) -- evaluate body in the environment
; extended with the bindings of id ... to the corresponding
; eigen-variables
(define-syntax eigen
  (syntax-rules ()
    ((_ (id ...) body)
      (let ((id (eigen-variable 'id)) ...) body))))

(test-check 'eigen
  (and
    (eigen () #t)
    (eigen (x) (eigen-var? x))
    (eigen (x y)
      (begin (display "eigens: ") (display (list x y))
	(newline) #t)))
  #t)

;;; ------------------------------------------------------

(define commitment cons)             ;;; change to list
(define commitment->term cdr)        ;;; and change to cadr
(define commitment->var car)

(define empty-subst '())
(define empty-subst? null?)

(define unit-subst
  (lambda (var t)
    (if (and (var? t) (eq? t _)) empty-subst (list (commitment var t)))))

(define extend-subst
  (lambda (unbound-var contains-no-bound-vars subst)
    (cons (commitment unbound-var contains-no-bound-vars) subst)))

(define cons-if-real-commitment
  (lambda (var term subst)
    (cond
      [(eq? term var) subst]
      [else (cons (commitment var term) subst)])))

; get the free vars of a term (a list without duplicates)
(define vars-of
  (lambda (term)
    (let loop ([term term] [fv '()])
      (cond
        [(var? term) (if (memq term fv) fv (cons term fv))]
        [(pair? term) (loop (cdr term) (loop (car term) fv))]
        [else fv]))))

; Check to see if a var occurs in a term
(define occurs?
  (lambda (var term)
    (cond
      [(var? term) (eq? term var)]
      [(pair? term) (or (occurs? var (car term)) (occurs? var (cdr term)))]
      [else #f])))


(define compose-subst/own-survivors
  (lambda (base refining survivors)
    (let refine ([b* base])
      (if (null? b*) survivors
          (cons-if-real-commitment
            (commitment->var (car b*))
            (subst-in (commitment->term (car b*)) refining)
            (refine (cdr b*)))))))

(define compose-subst
  (lambda (base refining)
    (cond
      [(null? base) refining]
      [(null? refining) base]
      [else
        (compose-subst/own-survivors base refining
          (let survive ([r* refining])
            (cond
              [(null? r*) '()]
              [(assq (commitment->var (car r*)) base) (survive (cdr r*))]
              [else (cons (car r*) (survive (cdr r*)))])))])))

; Replace a logical variable with the corresponding eigen-variable
(define universalize
  (lambda (term)
    (let ([fv (vars-of term)])
      (let ([subst
              (map
                (lambda (v)
                  (commitment v (eigen-variable (logical-variable-id v))))
                fv)])
        (subst-in term subst)))))

; Similar to universalize: makes nicer symbols for variables that look
; nicer when printed. The 'reverse' in (reverse (vars-of t))
; just makes the output look as it used to look before. Consider it
; a historical accident.
(define concretize
  (lambda (t)
    (subst-in-deep t
      (let loop ([fv (reverse (vars-of t))] [env '()])
	(cond
	  [(null? fv) empty-subst]
	  [else (let ([id (logical-variable-id (car fv))])
		  (let ([num (let*-and 0 ([pr (assq id env)]) (+ (cdr pr) 1))])
		    (cons (commitment (car fv) (artificial-id id num))
		      (loop (cdr fv) (cons (cons id num) env)))))])))))

(define subst-in-deep ; it will later become just subst-in...
  (lambda (t subst)
    (cond
      [(var? t)
       (let ([c (assq t subst)])
         (if c (subst-in-deep (commitment->term c) subst) t))]
      [(pair? t)
       (cons
         (subst-in-deep (car t) subst)
         (subst-in-deep (cdr t) subst))]
      [else t])))

(define artificial-id
  (lambda (t-id num)
    (string->symbol
      (string-append
        (symbol->string t-id) "." (number->string num)))))

(define-syntax concretize-subst/vars
  (syntax-rules ()
    [(_ subst) '()]
    [(_ subst x ... )
      (concretize (list (list x (subst-in x subst)) ...))]))


(define _ (let-lv (_) _))

(define concretize-subst concretize)

(define assq/var-id
  (lambda (id env)
    (cond
      [(null? env) #f]
      [(eq? (logical-variable-id (caar env)) id) (car env)]
      [else (assq/var-id id (cdr env))])))


;------------------------------------------------------------------------
;;; This is the unifier of Oleg Kiselov
;;; These two definitions are for displaying purposes, only
;;; They have nothing to do with unification.

(define normalize-subst
  (lambda (subst)
    (map (lambda (c)
           (commitment (commitment->var c)
             (subst-vars-recursively (commitment->term c) subst)))
      subst)))

(define subst-vars-recursively
  (lambda (t subst)
    (cond
      [(var? t)
       (cond
         [(assq t subst) =>
          (lambda (c)
            (subst-vars-recursively
              (commitment->term c) (remq c subst)))]
         [else t])]
      [(pair? t)
       (cons
         (subst-vars-recursively (car t) subst)
         (subst-vars-recursively (cdr t) subst))]
      [else t])))

;;;; ------------------------------------------------------

;;;; This new subst-in is essential to Oleg's unifier.
(define subst-in
  (lambda (t subst)
    (cond
      [(var? t)
       (let ([c (assq t subst)])
         (if c (subst-in (commitment->term c) subst) t))]
      [(pair? t)
       (cons
         (subst-in (car t) subst)
         (subst-in (cdr t) subst))]
      [else t])))

;;;; This is Oleg's unifier

; Either t or u may be:
; _
; free-var
; bound-var
; pair
; other-value
; So, we have in general 25 possibilities to consider.
; actually, a pair or components of a pair can be variable-free
; or not. In the latter case, we have got to traverse them.
; Also, if a term to unify has come from subst, it has special properties,
; which we can exploit. See below.
;
; "Measurements of the dynamic behavior of unification on four real
; programs show that one or both of the arguments are variables about
; 85% of the time [63]. A subroutine call is made only if both arguments
; are nonvariables." (Peter Van Roy, The Wonder Years ...)
;
; Just like in the union-find unification algorithm, we produce
; substitutions in the "triangular form" (see Baader, Snyder, Unification
; Theory). Circularity is detected only at the end (when we do subst-in).

(define unify
  (lambda (t u subst)
    (cond
      [(eq? t u) subst]			; quick tests first
      [(eq? t _) subst]
      [(eq? u _) subst]
      [(var? t)
       (let*-and (unify-free/any t u subst) ([ct (assq t subst)])
	 (if (var? u)			; ct is a bound var, u is a var
	   (let*-and (unify-free/bound u ct subst) ([cu (assq u subst)])
	     (unify-bound/bound ct cu subst))
	   (unify-bound/nonvar ct u subst)))]
      [(var? u)				; t is not a variable...
       (let*-and
         (cond
           [(pair? t) (unify-free/list u t subst)]
           ; t is not a var and is not a pair: it's atomic
           [else (extend-subst u t subst)])
         ([cu (assq u subst)])
         (unify-bound/nonvar cu t subst))]
      [(and (pair? t) (pair? u))
       (let*-and #f ([subst (unify (car t) (car u) subst)])
         (unify (cdr t) (cdr u) subst))]
      [else (and (equal? t u) subst)])))

; ct is a commitment to a bound variable, u is a atomic or a composite
; value -- but not a variable
(define unify-bound/nonvar
  (lambda (ct u subst)
    (let ((t (commitment->term ct)))
      (cond				; search for the end of ct -> chain
	[(eq? t u) subst]
	[(var? t)
	  (let*-and 
	    (cond
	      [(pair? u) (unify-free/list t u subst)]
              ; u is not a var and is not a pair: it's atomic
	      [else (extend-subst t u subst)])
	    ([ct (assq t subst)])
	    (unify-bound/nonvar ct u subst))]
	; t is some simple or composite value. So is u.
      [(and (pair? t) (pair? u))
	(let*-and #f ([subst (unify-internal/any (car t) (car u) subst)])
	  (unify-internal/any (cdr t) (cdr u) subst))]
      [else (and (equal? t u) subst)]))))


; Just like unify. However, the first term, t, comes from
; an internalized term. We know it can't be _ and can't contain _

(define unify-internal/any
  (lambda (t u subst)
    (cond
      [(eq? t u) subst]			; quick tests first
      [(eq? u _) subst]
      [(var? t)
       (let*-and (unify-free/any t u subst) ([ct (assq t subst)])
	 (if (var? u)			; ct is a bound var, u is a var
	   (let*-and (unify-free/bound u ct subst) ([cu (assq u subst)])
	     (unify-bound/bound ct cu subst))
	   (unify-bound/nonvar ct u subst)))]
      [(var? u)				; t is not a variable...
       (let*-and			; It's a part of an internal term
	 (extend-subst u t subst)	; no further checks needed
         ([cu (assq u subst)])
         (unify-internals (commitment->term cu) t subst))]
      [(and (pair? t) (pair? u))
       (let*-and #f ([subst (unify-internal/any (car t) (car u) subst)])
         (unify-internal/any (cdr t) (cdr u) subst))]
      [else (and (equal? t u) subst)])))


; Unify two already bound variables represented by their commitments
; ct and cu.
; We single out this case because in the future we may wish
; to unify the classes of these variables, by making a redundant
; binding of (commitment->var ct) to (commitment->term cu) or
; the other way around.
; Aside from the above, this function can take advantage of the following
; facts about (commitment->term cx) (where cx is an existing commitment):
;   - it is never _
;   - it never contains _
; Most importantly, if, for example, (commitment->term ct) is a free variable,
; we enter its binding to (commitment->term cu) with fewer checks.
; in particular, we never need to call unify-free/list nor
; unify-free/any as we do need to rebuild any terms.

(define unify-internals
  (lambda (t u subst)
      (cond
	[(eq? t u) subst]               ; quick tests first
	[(var? t)
         (let*-and (cond                ; t is a free variable
                     [(var? u)
                      (let*-and (extend-subst t u subst) ([cu (assq u subst)])
                        (unify-free/bound t cu subst))]
                     [else              ; t is free, u is not a var: done
                       (extend-subst t u subst)])
           ([ct (assq t subst)])
           (cond			; t is a bound variable
             [(var? u) 
              (let*-and (unify-free/bound u ct subst) ([cu (assq u subst)])
                (unify-bound/bound ct cu subst))]
             [else                      ; unify bound and a value
               (unify-internals (commitment->term ct) u subst)]))]
	[(var? u)                       ; t is not a variable...
         (let*-and (extend-subst u t subst) ([cu (assq u subst)])
           (unify-internals (commitment->term cu) t subst))]
        [(and (pair? t) (pair? u))
         (let*-and #f ([subst (unify-internals (car t) (car u) subst)])
           (unify-internals (cdr t) (cdr u) subst))]
        [else (and (equal? t u) subst)])))

(define unify-bound/bound
  (lambda (ct cu subst)
    (unify-internals (commitment->term ct) (commitment->term cu) subst)))


; t-var is a free variable, u can be anything
; This is analogous to get_variable instruction of Warren Abstract Machine
; (WAM).
; This function is not recursive and always succeeds, 
; because unify-free/bound and unify-free/list always succeed.
(define unify-free/any
  (lambda (t-var u subst)
    (cond
      [(eq? u _) subst]
      [(var? u)
       (let*-and (extend-subst t-var u subst) ([cu (assq u subst)])
         (unify-free/bound t-var cu subst))]
      [(pair? u) (unify-free/list t-var u subst)]
      [else ; u is not a var and is not a pair: it's atomic
	(extend-subst t-var u subst)])))

; On entrance: t-var is free.
; we are trying to unify it with a bound variable (commitment->var cu)
; Chase the binding chain, see below for comments
; This also works somewhat like union-find...
; This function always succeeds. The resulting substitution is either
; identical to the input one, or differs only in the binding to t-var.
;
; Unlike the previous version of the unifier,
; The following code does not introduce the temp variables *a and *d
; It makes substitutions more complex. Therefore, pruning them
; will take a while, and will break up the sharing. Therefore, we
; don't do any pruning.

(define unify-free/bound
  (lambda (t-var cu s)
    (let loop ([cm cu])
      (let ([u-term (commitment->term cm)])
	(cond
	  [(eq? u-term t-var) s]
	  [(var? u-term)
	    (cond
	      [(assq u-term s) => loop]
	      [else (extend-subst t-var u-term s)])] ; u-term is free here
	  [else (extend-subst t-var u-term s)])))))

; ((and (pattern-var? tree2) (assq tree2 env)) => ; tree2 is a bound var
;        ; binding a free variable to a bound. Search for a substantial binding
;        ; or a loop. If we find a loop tree1->tree2->...->tree1
;        ; then we do not enter the binding to tree1, because tree1 is not
;        ; actually constrained.
;       (lambda (tree2-binding)
; 	(let loop ((binding tree2-binding))
; 	  (cond
; 	    ((eq? tree1 (cdr binding)) env)  ; loop: no binding needed
; 	    ((and (pattern-var? (cdr binding)) (assq (cdr binding) env))
; 	      => loop)
; 	    (else (cons (cons tree1 (cdr binding)) env))))))

(define ground?
  (lambda (t)
    (cond
      [(var? t) #f]
      [(pair? t) (and (ground? (car t)) (ground? (cdr t)))]
      [else #t])))

; t-var is a free variable, u-value is a proper or improper
; list, which may be either fully or partially grounded (or not at all).
; We scan the u-value for _, and if, found, replace them with fresh
; variables. We then bind t-var to the term.
; This function is not recursive and always succeeds.
;
; We assume that more often than not u-value does not contain _.
; Therefore, to avoid the wasteful rebuilding of u-value, we 
; first scan it for the occurrence of _. If the scan returns negative,
; we can use u-value as it is.

      ; Rebuild lst replacing all anonymous variables with some
      ; fresh logical variables
      ; If lst contains no anonymous variables, return #f
      ; Note that lst itself may be #f -- and yet no contradiction arises.
(define ufl-rebuild-without-anons
  (lambda (lst)
    (cond
      [(eq? lst _) (logical-variable '*anon)]
      [(not (pair? lst)) #f]
      [(null? (cdr lst))
	(let [(new-car (ufl-rebuild-without-anons (car lst)))]
	  (and new-car (cons new-car '())))]
      [else
	(let [(new-car (ufl-rebuild-without-anons (car lst)))
              (new-cdr (ufl-rebuild-without-anons (cdr lst)))]
	  (if new-car
	    (cons new-car (or new-cdr (cdr lst)))
	    (and new-cdr (cons (car lst) new-cdr))))])))

(define unify-free/list
  (lambda (t-var u-value subst)
    (extend-subst t-var
      (or (ufl-rebuild-without-anons u-value) u-value)
      subst)))

(let-lv (x y)
  (test-check 'test-compose-subst-0
    (append (unit-subst x y) (unit-subst y 52))
    `(,(commitment x y) ,(commitment y 52))))


(test-check 'test-compose-subst-1
  (let-lv (x y)
    (equal?
      (compose-subst (unit-subst x y) (unit-subst y 52))
      `(,(commitment x 52) ,(commitment y 52))))
  #t)

(test-check 'test-compose-subst-2
  (let-lv (w x y)
    (equal?
      (let ([s (compose-subst (unit-subst y w) (unit-subst w 52))])
	(compose-subst (unit-subst x y) s))
      `(,(commitment x 52) ,(commitment y 52) ,(commitment w 52))))
  #t)

'(let-lv (w x y)
  (test-check 'test-compose-subst-3
    (let ([s (compose-subst (unit-subst w 52) (unit-subst y w))])
      (compose-subst (unit-subst x y) s))
    `(,(commitment x w) ,(commitment w 52) ,(commitment y w))))

'(test-check 'test-compose-subst-4
  (let-lv (x y z)
    (equal?
      (let ([s (compose-subst (unit-subst y z) (unit-subst x y))]
	    [r (compose-subst
		 (compose-subst (unit-subst x 'a) (unit-subst y 'b))
		 (unit-subst z y))])
	(compose-subst s r))
      `(,(commitment x 'b) ,(commitment z y))))
  #t)

(test-check 'test-compose-subst-5
  (concretize-subst
    (compose-subst
      (let-lv (x) (unit-subst x 3))
      (let-lv (x) (unit-subst x 4))))
  '((x.0 . 3) (x.1 . 4)))


;  Fk  = () -> Ans
;  Ans =  Nil + [Subst,Fk]
;  Sk = Fk -> Subst -> Ans  
;  Antecedent = Sk -> Sk
;  Rule = Antecedent -> [Goal-fn,Int] -> Antecedent

;  relation: Term -> Antecedent
;  to-show: Term -> Antecedent -> Rule
;  initial-sk : Sk
;  initial-fk : Fk

(define initial-fk (lambda () '()))
(define initial-sk (lambda@ (fk subst)
		     (cons subst fk)))

;------------------------------------------------------------------------
; Making logical variables "scoped" and garbage-collected

; A framework to remove introduced variables when they leave their scope.
; To make removing variables easier, we consider the list of subst as a
; "stack". Before we add a new variable, we retain a pointer to the
; stack. Then, when we are about to remove the added variables after their
; scope is ended, we stop at the shared retained substitution, and we know
; that anything below the retained substitution can't possibly contain the
; reference to the variables we're about to remove.
;
; Pruning of substitutions is analogous to environment pruning (aka tail-call
; optimization) in WAM on _forward_ execution.

; LV-ELIM IN-SUBST SUBST ID ....
; remove the bindings of ID ... from SUBST (by composing with the
; rest of subst). IN-SUBST is the mark.
; If we locate IN-SUBST in SUBST, we know that everything below the
; mark can't possibly contain ID ...

; lv-elim-1 VAR IN-SUBST SUBST
; VAR is a logical variable, SUBST is a substitution, and IN-SUBST
; is a tail of SUBST (which may be '()).
; VAR is supposed to have non-complex binding in SUBST
; (see Definition 3 in the document "Properties of Substitutions").
; If VAR is bound in SUBST, the corresponding commitment 
; is supposed to occur in SUBST up to but not including IN-SUBST.
; According to Proposition 10, if VAR freely occurs in SUBST, all such
; terms are VAR itself.
; The result is a substitution with the commitment to VAR removed
; and the other commitments composed with the removed commitment.
; The order of commitments is preserved.

(define lv-elim-1
  (lambda (var in-subst subst)
    (if (eq? subst in-subst) subst
      ; if VAR is not bound, there is nothing to prune
      (let*-and subst ([var-binding (assq var subst)])
	(let ([tv (commitment->term var-binding)])
	  (let loop ([current subst])
	    (cond
	      [(null? current) current]
	      [(eq? current in-subst) current]
	      [(eq? (car current) var-binding)
	       (loop (cdr current))]
	      [(eq? (commitment->term (car current)) var)
	       (cons (commitment (commitment->var (car current)) tv)
		 (loop (cdr current)))]
	      [else (cons (car current) (loop (cdr current)))])))))))

; The same but for multiple vars
; To prune multiple-vars, we can prune them one-by-one
; We can attempt to be more efficient and prune them in parallel.
; But we encounter a problem:
; If we have a substitution
;  ((x . y) (y . 1) (a . x))
; Then pruning 'x' first and 'y' second will give us ((a . 1))
; Pruning 'y' first and 'x' second will give us ((a . 1))
; But naively attempting to prune 'x' and 'y' in parallel
; disregarding dependency between them results in ((a . y))
; which is not correct.
; We should only be concerned about a direct dependency:
;  ((x . y) (y . (1 t)) (t . x) (a . x))
; pruning x and y in sequence or in parallel gives the same result:
;  ((t . (1 t)) (a . (1 t)))
; We should also note that the unifier will never return a substitution
; that contains a cycle ((x1 . x2) (x2 . x3) ... (xn . x1))

(define lv-elim
  (lambda (vars in-subst subst)
    (if (eq? subst in-subst)
      subst
      (let ([var-bindings ; the bindings of truly bound vars
	      (let loop ([vars vars])
		(if (null? vars) vars
		  (let ([binding (assq (car vars) subst)])
		    (if binding
		      (cons binding (loop (cdr vars)))
		      (loop (cdr vars))))))])
	(cond
	  [(null? var-bindings) subst] ; none of vars are bound
	  [(null? (cdr var-bindings))
	    ; only one variable to prune, use the faster version
	   (lv-elim-1 (commitment->var (car var-bindings))
	     in-subst subst)]
	  [(let test ([vb var-bindings]) ; check multiple dependency
	     (and (pair? vb)
	       (or (let ([term (commitment->term (car vb))])
		     (and (var? term) (assq term var-bindings)))
		 (test (cdr vb)))))
	    ; do pruning sequentially
	   (let loop ([var-bindings var-bindings] [subst subst])
	     (if (null? var-bindings) subst
	       (loop (cdr var-bindings)
		 (lv-elim-1 (commitment->var (car var-bindings))
		   in-subst subst))))]
	  [else				; do it in parallel
	    (let loop ([current subst])
	      (cond
		[(null? current) current]
		[(eq? current in-subst) current]
		[(memq (car current) var-bindings)
		 (loop (cdr current))]
		[(assq (commitment->term (car current)) var-bindings) =>
		 (lambda (ct)
		   (cons (commitment (commitment->var (car current)) 
			   (commitment->term ct))
		     (loop (cdr current))))]
		[else (cons (car current) (loop (cdr current)))]))])))))

; when the unifier is moved up, move lv-elim test from below up...

; That was the code for the unifier that introduced temp variables
; (define-syntax exists
;   (syntax-rules ()
;     [(_ () ant) ant]
;     [(_ (ex-id) ant)
;      (let-lv (ex-id)
;        (lambda@ (sk fk in-subst)
;          (@ ant
;            (lambda@ (fk out-subst)
;              (@ sk fk (lv-elim-1 ex-id in-subst out-subst)))
;            fk in-subst)))]
;     [(_ (ex-id ...) ant)
;      (let-lv (ex-id ...) 
;        (lambda@ (sk fk in-subst)
;          (@ ant
;            (lambda@ (fk out-subst)
;              (@ sk fk (lv-elim (list ex-id ...) in-subst out-subst)))
;            fk in-subst)))]))

; For the unifier that doesn't introduce temp variables,
; exists is essentially let-lv
; At present, we don't do any GC.
; Here's the reason we don't do any pruning now:
; Let's unify the variable x with a term `(1 2 3 4 5 ,z). The result
; will be the binding x -> `(1 2 3 4 5 ,z). Let's unify `(1 . ,y) with
; x. The result will be a binding y -> `(2 3 4 5 ,z). Note that the
; bindings of x and y share a tail. Let us now unify z with 42. The
; result will be a binding z->42. So far, so good. Suppose however that
; z now "goes out of scope" (the exists form that introduced z
; finishes). We now have to traverse all the terms in the substitution
; and replace z with its binding. The result will be a substitution
; 	x -> (1 2 3 4 5 42)
; 	y -> (2 3 4 5 42)
; Now, the bindings of x and y do not share anything at all! The pruning
; has broke sharing. If we want to unify x and `(1 . ,y) again, we have
; to fully traverse the corresponding terms again.
; So, to prune variables and preserve sharing, we have to topologically sort
; the bindings first!

(define-syntax exists
  (syntax-rules ()
    [(_ () ant) ant]
    [(_ (ex-id ...) ant)
     (let-lv (ex-id ...) ant)]
    ))

;-----------------------------------------------------------
; Sequencing of relations
; Antecedent is a multi-valued function (which takes
;   sk, fk, subst and exits to either sk or fk).
; A relation is a parameterized antecedent.
;
; All sequencing operations are defined on antecedents.
; They can be "lifted" to relations (see below).
; 

; TRACE-ANT-RAW TITLE ANT -> ANT
; Traces all invocations and re-invocations of an antecedent
; printing subst before and after, in their raw form
(define trace-ant-raw
  (lambda (title ant)
    (let ((print-it
	    (lambda (event subst)
	      (display title) (display " ")
	      (display event) (pretty-print subst) (newline))))
      (lambda@ (sk fk subst)
	(print-it "CALL:" subst)
	(@ ant 
	  (lambda@ (fk subst)
	    (print-it "RETURN:" subst)
	    (@ sk
	      (lambda ()
		(display title) (display " REDO") (newline)
		(fk))
	      subst))
	  (lambda ()
	    (display title) (display " FAIL") (newline)
	    (fk))
	  subst)))))

; Conjunctions
; All conjunctions below satisfy properties
;    ans is an answer of (a-conjunction ant1 ant2 ...) ==>
;       forall i. ans is an answer of ant_i
;    (a-conjunction) ==> true


; (all ant1 ant2 ...)
; A regular Prolog conjunction. Non-deterministic (i.e., can have 0, 1,
; or more answers).
; Properties:
;  (all ant) ==> ant
;  (all ant1 ... ant_{n-1} antn) is a "join" of answerlists of
;        (all ant1 ... ant_{n-1}) and antn

(define-syntax all
  (syntax-rules ()
    [(_) (lambda@ (sk) sk)]
    [(_ ant) ant]
    [(_ ant0 ant1 ...)
     (lambda@ (sk)
       (splice-in-ants/all sk ant0 ant1 ...))]))

(define-syntax splice-in-ants/all
  (syntax-rules ()
    [(_ sk ant) (@ ant sk)]
    [(_ sk ant0 ant1 ...)
     (@ ant0 (splice-in-ants/all sk ant1 ...))]))

(define succeed (all))

; (promise-one-answer ant)
; Operationally, it is the identity.
; It is an optimization directive: if the user knows that an antecedent
; can produce at most one answer, he can tell the system about it.
; The behavior is undefined if the user has lied.

(define-syntax promise-one-answer
  (syntax-rules ()
    [(_ ant) ant]))

; (all! ant1 ant2 ...)
; A committed choice nondeterminism conjunction
; From the Mercury documentation:

;   In addition to the determinism annotations described earlier, there
;   are "committed choice" versions of multi and nondet, called cc_multi
;   and cc_nondet. These can be used instead of multi or nondet if all
;   calls to that mode of the predicate (or function) occur in a context
;   in which only one solution is needed.
;
; (all! ant) evaluates ant in a single-choice context. That is,
; if ant fails, (all! ant) fails. If ant has at least one answer,
; this answer is returned.
; (all! ant) has at most one answer regardless of the answers of ant.
;   ans is an answer of (all! ant) ==> ans is an answer of ant
; The converse is not true.
; Corollary: (all! ant) =/=> ant
; Corollary: ant is (semi-) deterministic: (all! ant) ==> ant
; (all! (promise-one-answer ant)) ==> ant
;
; By definition, (all! ant1 ant2 ...) ===> (all! (all ant1 ant2 ...))

(define-syntax all!
  (syntax-rules (promise-one-answer)
    [(_) (promise-one-answer (all))]
    [(_ (promise-one-answer ant)) (promise-one-answer ant)] ; keep the mark
    [(_ ant0 ant1 ...)
     (promise-one-answer
       (lambda@ (sk fk)
	 (@
	   (splice-in-ants/all (lambda@ (fk-ign) (@ sk fk)) ant0 ant1 ...)
	   fk)))]))

; (all!! ant1 ant2 ...)
; Even more committed choice nondeterministic conjunction
; It evaluates all elements of the conjunction in a single answer context
; (all!! ant) ==> (all! ant) =/=> ant
; (all!! ant1 ant2 ...) ==> (all (all! ant1) (all! ant2) ...)
;                       ==> (all! (all! ant1) (all! ant2) ...)
; (all!! ant1 ... antn (promise-one-answer ant)) ==>
;    (all (all!! ant1 ... antn) ant)

(define-syntax all!!
  (syntax-rules ()
    [(_) (all!)]
    [(_ ant) (all! ant)]
    [(_ ant0 ant1 ...)
     (promise-one-answer 
       (lambda@ (sk fk)
         (splice-in-ants/all!! sk fk ant0 ant1 ...)))]))

(define-syntax splice-in-ants/all!!
  (syntax-rules (promise-one-answer)
    [(_ sk fk)
      (@ sk fk)]
    [(_ sk fk (promise-one-answer ant))
      (@ ant sk fk)]
    [(_ sk fk ant0 ant1 ...)
      (@ ant0 (lambda (fk-ign) (splice-in-ants/all!! sk fk ant1 ...)) fk)]))

; (if-only COND THEN)
; (if-only COND THEN ELSE)
; Here COND, THEN, ELSE are antecedents.
; If COND succeeds at least once, the result is equivalent to
;      (all (all! COND) TNEN)
; If COND fails, the result is the same as ELSE.
; If ELSE is omitted, it is assumed fail. That is, (if-only COND THEN)
; fails if the condition fails.  "This  unusual semantics
; is part of the ISO and all de-facto Prolog standards."
; Thus, declaratively,
;   (if-only COND THEN ELSE) ==> (any (all (all! COND) THEN)
;                                     (all (fails COND) ELSE))
; Operationally, we try to generate a good code.

; "The majority of predicates written by human programmers are
; intended to give at most one solution, i.e., they are
; deterministic. These predicates are in effect case statements
; [sic!], yet they are too often compiled in an inefficient manner
; using the full generality of backtracking (which implies saving the
; machine state and repeated failure and state restoration)." (Peter
; Van Roy, 1983-1993: The Wonder Years of Sequential Prolog
; Implementation).


(define-syntax if-only
  (syntax-rules ()
    [(_ condition then)
     (lambda@ (sk fk)
       (@ condition
         ; sk from cond
         (lambda@ (fk-ign) (@ then sk fk))
         ; failure from cond
         fk))]
    [(_ condition then else)
     (lambda@ (sk fk subst)
       (@ condition
         (lambda@ (fk-ign) (@ then sk fk))
         (lambda () (@ else sk fk subst))
         subst))]))

; (if-all! (COND1 ... CONDN) THEN)
; (if-all! (COND1 ... CONDN) THEN ELSE)
;
; (if-all! (COND1 ... CONDN) THEN ELSE) ==>
;   (if-only (all! COND1 ... CONDN) THEN ELSE)
; (if-all! (COND1) THEN ELSE) ==>
;   (if-only COND1 THEN ELSE)

; Eventually, it might be a recognized special case in if-only.

; (define-syntax if-all!
;   (syntax-rules ()
;     [(_ (condition) then) (if-only condition then)]
;     [(_ (condition) then else) (if-only condition then else)]
;     [(_ (condition1 condition2 ...) then)
;      (lambda@ (sk fk)
;        (@ (splice-in-ants/all
;             (lambda@ (fk-ign)
;               (@ then sk fk))
;             condition1 condition2 ...)
;           fk))]
;     [(_ (condition1 condition2 ...) then else)
;      (lambda@ (sk fk subst)
;        (@ (splice-in-ants/all
;             (lambda@ (fk-ign)
;               (@ then sk fk)) condition1 condition2 ...)
;           (lambda ()
;             (@ else sk fk subst))
; 	 subst))]))

; Disjunction of antecedents
; All disjunctions below satisfy properties
;  ans is an answer of (a-disjunction ant1 ant2 ...) ==>
;    exists i. ans is an answer of ant_i
; (a-disjunction) ==> fail

; Any disjunction. A regular Prolog disjunction (introduces
; a choicepoints, in Prolog terms)
; Note that 'any' is not a union! In particular, it is not
; idempotent.
; (any) ===> fail
; (any ant) ===> ant
; (any ant1 ... antn) ==> _concatenation_ of their answerlists

(define-syntax any
  (syntax-rules ()
    [(_) (lambda@ (sk fk subst) (fk))]
    [(_ ant) ant]
    [(_ ant ...)
      (lambda@ (sk fk subst)
	(splice-in-ants/any sk fk subst ant ...))]))

(define-syntax splice-in-ants/any
  (syntax-rules ()
    [(_ sk fk subst ant1) (@ ant1 sk fk subst)]
    [(_ sk fk subst ant1 ant2 ...)
     (@ ant1 sk (lambda ()
                  (splice-in-ants/any sk fk subst ant2 ...))
       subst)]))

(define fail (any))

; Negation
; (fails ant) succeeds iff ant has no solutions
; (fails ant) is a semi-deterministic predicate: it can have at most
; one solution
; (succeeds ant) succeeds iff ant has a solution
;
; (fails (fails ant)) <===> (succeeds ant)
; but (succeeds ant) =/=> ant
; Cf. (equal? (not (not x)) x) is #f in Scheme in general.

(define fails
  (lambda (ant)
    (lambda@ (sk fk subst)
      (@ ant
        (lambda@ (current-fk subst) (fk))
        (lambda () (@ sk fk subst))
        subst))))

(define succeeds
  (lambda (ant)
    (lambda@ (sk fk subst)
      (@ ant (lambda@ (fk-ign subst-ign) (@ sk fk subst))
	fk subst))))

; partially-eval-sant: Partially evaluate a semi-antecedent. A
; semi-antecedent is an expression that, when applied to two
; arguments, sk and fk, can produce zero, one, or more answers.  Any
; antecedent can be turned into a semi-antecedent if partially applied
; to subst.  The following higher-order semi-antecedent takes an
; antecedent and yields the first answer and another, residual
; antecedent. The latter, when evaluated, will give the rest of the
; answers of the original semi-antecedent.  partially-eval-sant could
; be implemented with streams (lazy lists). The following is a purely
; combinational implementation.
;
; (@ partially-eval-sant sant a b) =>
;   (b) if sant has no answers
;   (a s residial-sant) if sant has a answer. That answer is delivered
;                       in s. 
; The residial semi-antecedent can be passed to partially-eval-sant
; again, and so on, to obtain all answers from an antecedent one by one.

; The following definition is eta-reduced.

(define (partially-eval-sant sant)
  (@ sant
    (lambda@ (fk subst a b)
      (@ a subst 
	(lambda@ (sk1 fk1)
	  (@
	    (fk) 
	    ; new a
	    (lambda@ (sub11 x) (@ sk1 (lambda () (@ x sk1 fk1)) sub11))
	    ; new b
	    fk1))))
    (lambda () (lambda@ (a b) (b)))))

(define (ant->sant ant subst)
  (lambda@ (sk fk)
    (@ ant sk fk subst)))

; An interleaving disjunction.
; Declaratively, any-interleave is the same as any.
; Operationally, any-interleave schedules each component antecedent
; in round-robin. So, any-interleave is fair: it won't let an antecedent
; that produces infinitely many answers (such as repeat) starve the others.
; any-interleave introduces a breadth-first-like traversal of the
; decision tree.
; I seem to have seen a theorem that says that a _fair_ scheduling
; (like that provided by any-interleave) entails a minimal or well-founded
; semantics of a Prolog program.

(define-syntax any-interleave
  (syntax-rules ()
    [(_) fail]
    [(_ ant) ant]
    [(_ ant ...)
     (lambda@ (sk fk subst)
       (interleave sk fk
         (list (ant->sant ant subst) ...)))]))

; we treat sants as a sort of a circular list
(define interleave
  (lambda (sk fk sants)
    (cond
      [(null? sants) (fk)]		; all of the sants are finished
      [(null? (cdr sants))
      ; only one sants left -- run it through the end
       (@ (car sants) sk fk)]
      [else
        (let loop ([curr sants] [residuals '()])
	  ; check if the current round is finished
	  (if (null? curr) (interleave sk fk (reverse residuals))
	    (@
	      partially-eval-sant (car curr)
	      ; (car curr) had an answer
	      (lambda@ (subst residual)
	        (@ sk
	          ; re-entrance cont
		  (lambda () (loop (cdr curr) (cons residual residuals)))
		  subst))
	    ; (car curr) is finished - drop it, and try next
	    (lambda () (loop (cdr curr) residuals)))))])))

; An interleaving disjunction removing duplicates: any-union
; This is a true union of the constituent antecedents: it is fair, and
; it removes overlap in the antecedents to union, if any. Therefore,
;    (any-union ant ant) ===> ant
; whereas (any ant ant) =/=> ant
; because the latter has twice as many answers as ant.
;
; Any-union (or interleave-non-overlap, to be precise) is quite similar
; to the function interleave above. But now, the order of antecedents
; matters. Given antecedents ant1 ant2 ... antk ... antn,
; at the k-th step we try to partially-eval antk. If it yields an answer,
; we check if ant_{k+1} ... antn can be satisfied with that answer.
; If any of them does, we disregard the current answer and ask antk for
; another one. We maintain the invariant that
;  ans is an answer of (any-union ant1 ... antn) 
;  ===> exists i. ans is an answer of ant_i
;       && forall j>i. ans is not an answer of ant_j
; The latter property guarantees the true union.
; Note the code below does not check if answers of each individual
; antecedent are unique. It is trivial to modify the code so that
; any-union removes the duplicates not only among the antecedents but
; also within an antecedent. That change entails a run-time cost. More
; importantly, it breaks the property
; (any-union ant ant) ===> ant
; Only a weaker version, (any-union' ant ant) ===> (any-union' ant)
; would hold. Therefore, we do not make that change.

(define-syntax any-union
  (syntax-rules ()
    [(_) fail]
    [(_ ant) ant]
    [(_ ant ...)
     (lambda@ (sk fk subst)
       (interleave-non-overlap sk fk
         (list (cons (ant->sant ant subst) ant) ...)))]))

; we treat saants as a sort of a circular list
; Each element of saants is a pair (sant . ant)
; where ant is the original antecedent (needed for the satisfiability testing)
; and sant is the corresponding semi-antecedent or a 
; residual thereof.
(define interleave-non-overlap
  (lambda (sk fk saants)
    (let outer ([saants saants])
      (cond
        [(null? saants) (fk)]  ; all of the saants are finished
        [(null? (cdr saants))  ; only one ant is left -- run it through the end
	 (@ (caar saants) sk fk)]
        [else
	  (let loop ([curr saants]
                     [residuals '()])
            ; check if the current round is finished
	    (if (null? curr) (outer (reverse residuals))
                (@
                 partially-eval-sant (caar curr)
                  ; (caar curr) had an answer
                 (lambda@ (subst residual)
                  ; let us see now if the answer, subst, satisfies any of the
                  ; ants down the curr.
                   (let check ([to-check (cdr curr)])
                     (if (null? to-check) ; OK, subst is unique, give it to user
                         (@ sk
                           ; re-entrance cont
                           (lambda ()
                             (loop (cdr curr) 
                               (cons (cons residual (cdar curr)) residuals)))
                           subst)
                         (@ (cdar to-check)
                            ; subst was the answer to some other ant: check failed
                            (lambda@ (fk1 subst1) 
                              (loop (cdr curr) 
                                (cons (cons residual (cdar curr)) residuals)))
                            ; subst was not the answer: continue check
                            (lambda () (check (cdr to-check)))
                            subst))))
                 ; (car curr) is finished - drop it, and try next
                 (lambda () (loop (cdr curr) residuals)))))]))))

; Another if-then-else
; (if-some COND THEN)
; (if-some COND THEN ELSE)
; Here COND, THEN, ELSE are antecedents.
; If COND succeeds at least once, the result is equivalent to
;      (all COND TNEN)
; If COND fails, the result is the same as ELSE.
; If ELSE is omitted, it is assumed fail. That is, (if-some COND THEN)
; fails if the condition fails.  "This  unusual semantics
; is part of the ISO and all de-facto Prolog standards."
; Thus, declaratively,
;   (if-some COND THEN ELSE) ==> (any (all COND THEN)
;                                     (all (fails COND) ELSE))
; from which follows
;   (if-some COND THEN)              ==> (all COND THEN)
;   (if-some COND THEN fail)         ==> (all COND THEN)
; but
;   (if-some COND succeed ELSE)     =/=> (any COND ELSE)
;
; Other corollary:
;   (if-some COND THEN ELSE) ==> (if-only (fails COND) ELSE (all COND THEN))
;
; Operationally, we try to generate a good code.
;
; In Prolog, if-some is called a soft-cut (aka *->). In Mercury,
; if-some is the regular IF-THEN-ELSE.
;
; We can implement if-some with partially-eval-sant. Given a COND, we
; peel off one answer, if possible. If there is one, we then execute THEN
; passing it the answer and the fk from COND so that if THEN fails,
; it can obtain another answer. If COND has no answers, we execute
; ELSE. Again, we can do all that purely declaratively, without
; talking about introducing and destroying choice points.

(define-syntax if-some
  (syntax-rules ()
    [(_ condition then) (all condition then)]
    [(_ condition then else)
     (lambda@ (sk fk subst)
       (@ partially-eval-sant (ant->sant condition subst)
         (lambda@ (ans residual)
           (@ then sk
             ; then failed. Check to see if condition has another answer
             (lambda () (@ residual (@ then sk) fk))
             ans))
             ; condition failed
         (lambda () (@ else sk fk subst))))]))

; Relations...........................

; The current incremented unification of argument passing is quite similar to
; the compilation of argument unifications in WAM.

; relation (VAR ...) (to-show TERM ...) [ANT]
; Defines a relation of arity (length '(TERM ...)) with an optional body
; ANT. VAR ... are logical variables that are local to the relation, i.e.,
; appear in TERM or ANT. It's better to list as VAR ... only logical
; variables that appear in TERM. Variables that appear only in ANT should
; be introduced with exists. That makes their existential quantification
; clearer. Variables that appear in TERM are universally quantified.
;
; relation (head-let TERM ...) [ANT]
; See relation-head-let below.
;
; relation (ANNOT-VAR ...) (to-show TERM ...) [ANT]  (see remark below!)
; where ANNOT-VAR is either a simple VAR or (once VAR)
; where 'once' is a distingushed symbol. The latter form introduces
; a once-var, aka linear variable. A linear variable appears only once in
; TERM ... and only at the top level (that is, one and only one TERM
; in the to-show pattern contains ONCE-VAR, and that term is ONCE-VAR
; itself). In addition, ONCE-VAR must appear at most once in the body ANT.
; (Of course, then ONCE-VAR could be _, instead.)
; If these conditions are satisfied, we can replace a logical variable
; ONCE-VAR with a regular Scheme variable.

; Alternative notation:
; (relation (a c) (to-show term1 (once c) term2) body)
; Makes it easier to deal with. But it is unsatisfactory:
; to-show becomes a binding form...
;
; When ``compiling'' a relation, we now look through the
; (to-show ...) pattern for a top-level occurrence of the logical variable
; introduced by the relation. For example:
;	(relation (x y) (to-show `(,x . ,y) x) body)
; we notice that the logical variable 'x' occurs at the top-level. Normally we
; compile the relation like that into the following
;    (lambda (g1 g2)
;      (exists (x y)
;        (lambda@ (sk fk subst)
; 	 (let*-and (fk) ((subst (unify g1 `(,x . ,y)  subst))
; 			 (subst (unify g2 x subst)))
; 	     (@ body sk fk subst)))))
;
; However, that we may permute the order of 'unify g...' clauses
; to read
;    (lambda (g1 g2)
;      (exists (x y)
;        (lambda@ (sk fk subst)
; 	 (let*-and (fk) ((subst (unify x g2 subst))
; 			 (subst (unify g1 `(,x . ,y)  subst))
; 			 )
; 	     (@ body sk fk subst)))))
;
; We may further note that according to the properties of the unifier
; (see below), (unify x g2 subst) must always succeed, 
; because x is a fresh variable.
; Furthermore, the result of (unify x g2 subst) is either subst itself,
; or subst with the binding of x. Therefore, we can check if
; the binding at the top of (unify x g2 subst) is the binding to x. If
; so, we can remove the binding and convert the variable x from being logical
; to being lexical. Thus, we compile the relation as
;
;    (lambda (g1 g2)
;      (exists (x y)
;        (lambda@ (sk fk subst)
; 	 (let* ((subst (unify-free/any x g2 subst))
; 	        (fast-path? (and (pair? subst)
; 			         (eq? x (commitment->var (car subst)))))
; 	        (x (if fast-path? (commitment->term (car subst)) x))
; 	        (subst (if fast-path? (cdr subst) subst)))
; 	 (let*-and (fk) ((subst (unify g1 `(,x . ,y)  subst))
; 			 )
; 	     (@ body sk fk subst))))))
;
; The benefit of that approach is that we limit the growth of subst and avoid
; keeping commitments that had to be garbage-collected later.

; http://kanren.sf.net

'(define-syntax relation
  (syntax-rules (to-show head-let)
    [(_ (head-let head-term ...) ant)
     (relation-head-let (head-term ...) ant)]
    [(_ (head-let head-term ...))       ; not particularly useful without body
     (relation-head-let (head-term ...))]
    [(_ () (to-show x ...) ant)         ; pattern with no vars _is_ linear
     (relation-head-let (`,x ...) ant)]
    [(_ () (to-show x ...))             ; the same without body: not too useful
     (relation-head-let (`,x ...))]
    [(_ (ex-id ...) (to-show x ...) ant)  ; body present
     (relation (ex-id ...) () (x ...) (x ...) ant)]
    [(_ (ex-id ...) (to-show x ...))      ; no body
     (relation (ex-id ...) () (x ...) (x ...))]
    [(_ (ex-id ...) (var0 ...) (x0 x1 ...) xs ant ...)
     (relation (ex-id ...) (var0 ... g) (x1 ...) xs ant ...)]
    [(_ (ex-id ...) () () () ant)       ; no arguments (no head-tests)
      (lambda ()
        (exists (ex-id ...) ant))]
    [(_ (ex-id ...) (g ...) () (x ...)) ; no body
     (lambda (g ...)
       (exists (ex-id ...)
         (all!! (promise-one-answer (== g x)) ...)))]
    [(_ (ex-id ...) (g ...) () (x ...) ant) ; the most general
     (lambda (g ...)
       (exists (ex-id ...)
         (if-all! ((promise-one-answer (== g x)) ...) ant)))]))

(define-syntax relation
  (syntax-rules (to-show head-let once _)
    [(_ (head-let head-term ...) ant)
     (relation-head-let (head-term ...) ant)]
    [(_ (head-let head-term ...))	; not particularly useful without body
     (relation-head-let (head-term ...))]
    [(_ () (to-show term ...) ant)	; pattern with no vars _is_ linear
     (relation-head-let (`,term ...) ant)]
    [(_ () (to-show term ...))		; the same without body: not too useful
     (relation-head-let (`,term ...))]
    [(_ (ex-id ...) (to-show term ...) ant)  ; body present
     (relation "a" () () (ex-id ...) (term ...) ant)]
    [(_ (ex-id ...) (to-show term ...))      ; no body
     (relation "a" () () (ex-id ...) (term ...))]
    ; process the list of variables and handle annotations
    [(_ "a" vars once-vars ((once id) . ids) terms . ant)
     (relation "a" vars (id . once-vars) ids terms . ant)]
    [(_ "a" vars once-vars (id . ids) terms . ant)
     (relation "a" (id . vars) once-vars ids terms . ant)]
    [(_ "a" vars once-vars () terms . ant)
     (relation "g" vars once-vars () () () (subst) terms . ant)]
    ; generating temp names for each term in the head
    ; don't generate if the term is a variable that occurs in
    ; once-vars
    ; For _ variables in the pattern, generate unique names for the lambda
    ; parameters, and forget them
    ; also, note and keep track of the first occurrence of a term
    ; that is just a var (bare-var) 
    [(_ "g" vars once-vars (gs ...) gunis bvars bvar-cl (_ . terms) . ant)
     (relation "g" vars once-vars (gs ... anon) gunis
       bvars bvar-cl terms . ant)]
    [(_ "g" vars once-vars (gs ...) gunis bvars (subst . cls)
           (term . terms) . ant)
     (id-memv?? term once-vars 
       ; success continuation: term is a once-var
       (relation "g" vars once-vars (gs ... term) gunis bvars (subst . cls)
	 terms . ant)
       ; failure continuation: term is not a once-var
       (id-memv?? term vars
	 ; term is a bare var
	 (id-memv?? term bvars
	   ; term is a bare var, but we have seen it already: general case
	   (relation "g" vars once-vars  (gs ... g) ((g . term) . gunis) 
	     bvars (subst . cls) terms . ant)
	   ; term is a bare var, and we have not seen it
	   (relation "g" vars once-vars (gs ... g) gunis
	     (term . bvars)
	     (subst
	       (subst (unify-free/any term g subst))
	       (fast-path? (and (pair? subst)
			      (eq? term (commitment->var (car subst)))))
	       (term (if fast-path? (commitment->term (car subst)) term))
	       (subst (if fast-path? (cdr subst) subst))
	       . cls)
	     terms . ant))
	 ; term is not a bare var
	 (relation "g" vars once-vars  (gs ... g) ((g . term) . gunis) 
	   bvars (subst . cls) terms . ant)))]
    [(_ "g" vars once-vars gs gunis bvars bvar-cl () . ant)
     (relation "f" vars once-vars gs gunis bvar-cl . ant)]

    ; Final: writing the code
    [(_ "f" vars () () () (subst) ant)   ; no arguments (no head-tests)
      (lambda ()
	(exists vars ant))]
                                    ; no tests but pure binding
    [(_ "f" (ex-id ...) once-vars (g ...) () (subst) ant)
     (lambda (g ...)
       (exists (ex-id ...) ant))]
				    ; the most general
    [(_ "f" (ex-id ...) once-vars (g ...) ((gv . term) ...) 
       (subst let*-clause ...) ant ...)
     (lambda (g ...)
       (exists (ex-id ...)
	 (lambda@ (sk fk subst)
	   (let* (let*-clause ...)
	     (let*-and (fk) ((subst (unify gv term subst)) ...)
	       (@ ant ... sk fk subst))))))]))

; A macro-expand-time memv function for identifiers
;	id-memv?? FORM (ID ...) KT KF
; FORM is an arbitrary form or datum, ID is an identifier.
; The macro expands into KT if FORM is an identifier that occurs
; in the list of identifiers supplied by the second argument.
; Otherwise, id-memv?? expands to KF.
; All the identifiers in (ID ...) must be unique.
; Two identifiers match if both refer to the same binding occurrence, or
; (both are undefined and have the same spelling).

(define-syntax id-memv??
  (syntax-rules ()
    ((id-memv?? form (id ...) kt kf)
      (let-syntax
	((test
	   (syntax-rules (id ...)
	     ((test id _kt _kf) _kt) ...
	     ((test otherwise _kt _kf) _kf))))
	(test form kt kf)))))

; Test cases
; (id-memv?? x (a b c) #t #f)
; (id-memv?? a (a b c) 'OK #f)
; (id-memv?? () (a b c) #t #f)
; (id-memv?? (x ...) (a b c) #t #f)
; (id-memv?? "abc" (a b c) #t #f)
; (id-memv?? x () #t #f)
; (let ((x 1))
;   (id-memv?? x (a b x) 'OK #f))
; (let ((x 1))
;   (id-memv?? x (a x b) 'OK #f))
; (let ((x 1))
;   (id-memv?? x (x a b) 'OK #f))


; relation-head-let (head-term ...) ant 
; A simpler, and more efficient kind of relation. The simplicity comes
; from a simpler pattern at the head of the relation. The pattern must
; be linear and shallow with respect to introduced variables.  The ant
; is optional (although omitting it doesn't make much sense in
; practice) There are two kinds of head-terms.  One kind is an
; identifier. This identifier is taken to be a logical identifier, to
; be unified with the corresponding actual argument.  Each logical
; identifier must occur exactly once.  Another kind of a head-terms is
; anything else. That anything else may be a constant, a scheme
; variable, or a complex term that may even include logical variables
; such as _ -- but not logical variables defined in the same head-let
; pattern.  To make the task of distinguishing logical identifiers
; from anything else easier, we require that anything else of a sort
; of a manifest constant be explicitly quoted or quasiquoted. It would
; be OK to add `, to each 'anything else' term.
;
; Examples:
; (relation-head-let (x y z) (foo x y z))
; Here x y and z are logical variables.
; (relation-head-let (x y '7) (foo x y))
; Here we used a manifest constant that must be quoted
; (relation-head-let (x y `(1 2 . ,_)) (foo x y))
; We used a quasi-quoted constant with an anonymous variable.
; (let ((z `(1 2 . ,_))) (relation-head-let (x y `,z) (foo x y))
; The same as above, but using a lexical Scheme variable.
; The binding procedure is justified by Proposition 9 of
; the Properties of Substitutions.
;
; 'head-let' is an example of "compile-time" simplifications. 
; For example, we distinguish constants in the term head at
; "compile time" and so we re-arrange the argument-passing
; unifications to handle the constants first.
; The test for the anonymous variable (eq? gvv0 _) below
; is an example of a global simplification with a run-time
; test. A compiler could have inferred the result of the test -- but only
; upon the global analysis of all the clauses.
; Replacing a logical variable with an ordinary variable, which does 
; not have to be pruned, is equivalent to the use of temporary and
; unsafe variables in WAM.

(define-syntax relation-head-let
  (syntax-rules ()
    [(_ (head-term ...) . ants)
     (relation-head-let "g" () (head-term ...) (head-term ...) . ants)]
    ; generate names of formal parameters
    [(_ "g" (genvar ...) ((head-term . tail-term) . ht-rest)
       head-terms . ants)
     (relation-head-let "g" (genvar ... g) ht-rest head-terms . ants)]
    [(_ "g" (genvar ...) (head-term . ht-rest) head-terms . ants)
     (relation-head-let "g" (genvar ... head-term) ht-rest head-terms . ants)]
    [(_ "g" genvars  () head-terms . ants)
     (relation-head-let "d" () () genvars head-terms genvars . ants)]
    ; partition head-terms into vars and others
    [(_ "d" vars others (gv . gv-rest) ((hth . htt) . ht-rest) gvs . ants)
     (relation-head-let "d" vars ((gv (hth . htt)) . others)
       gv-rest ht-rest gvs . ants)]
    [(_ "d" vars others (gv . gv-rest) (htv . ht-rest) gvs . ants)
     (relation-head-let "d" (htv . vars) others
       gv-rest ht-rest gvs . ants)]
    [(_ "d" vars others () () gvs . ants)
     (relation-head-let "f" vars others gvs . ants)]
 
    ; final generation
    [(_ "f" vars ((gv term) ...) gvs) ; no body
     (lambda gvs                                     ; don't bother bind vars
       (lambda@ (sk fk subst)
	 (let*-and (fk) ([subst (unify gv term subst)] ...)
	   (@ sk fk subst))))]

    [(_ "f" (var0 ...) ((gvo term) ...) gvs ant)
     (lambda gvs
       (lambda@ (sk fk subst)			; first unify the constants
	 (let*-and (fk) ([subst (unify gvo term subst)] ...)
           (let ([var0 (if (eq? var0 _) (logical-variable '?) var0)] ...)
             (@ ant sk fk subst)))))]))

; (define-syntax relation/cut
;   (syntax-rules (to-show)
;     [(_ cut-id (ex-id ...) (to-show x ...) ant ...)
;      (relation/cut cut-id (ex-id ...) () (x ...) (x ...) ant ...)]
;     [(_ cut-id ex-ids (var ...) (x0 x1 ...) xs ant ...)
;      (relation/cut cut-id ex-ids (var ... g) (x1 ...) xs ant ...)]
;     [(_ cut-id (ex-id ...) (g ...) () (x ...) ant ...)
;      (lambda (g ...)
;        (exists (ex-id ...)
;          (all! (== g x) ...
;            (lambda@ (sk fk subst cutk)
;              (let ([cut-id (!! cutk)])
;                (@ (all ant ...) sk fk subst cutk))))))]))

(define-syntax fact
  (syntax-rules ()
    [(_ (ex-id ...) term ...)
     (relation (ex-id ...) (to-show term ...))]))

; Lifting from antecedents to relations
; (define-rel-lifted-comb rel-syntax ant-proc-or-syntax)
; Given (ant-proc-or-syntax ant ...)
; define 
; (rel-syntax (id ...) rel-exp ...)
; We should make rel-syntax behave as a CBV function, that is,
; evaluate rel-exp early.
; Otherwise, things like
; (define father (extend-relation father ...))
; loop.

; (define-syntax extend-relation
;   (syntax-rules ()
;     [(_ (id ...) rel-exp ...)
;      (extend-relation-aux (id ...) () rel-exp ...)]))

; (define-syntax extend-relation-aux
;   (syntax-rules ()
;     [(_ (id ...) ([g rel-exp] ...))
;      (let ([g rel-exp] ...)
;        (lambda (id ...)
;          (any (g id ...) ...)))]
;     [(_ (id ...) (let-pair ...) rel-exp0 rel-exp1 ...)
;      (extend-relation-aux (id ...)
;        (let-pair ... [g rel-exp0]) rel-exp1 ...)]))

(define-syntax define-rel-lifted-comb
  (syntax-rules ()
    [(_ rel-syntax-name ant-proc-or-syntax)
     (define-syntax rel-syntax-name
       (syntax-rules ()
         [(_ ids . rel-exps)
          (lift-ant-to-rel-aux ant-proc-or-syntax ids () . rel-exps)]))]))

(define-syntax lift-ant-to-rel-aux
  (syntax-rules ()
    [(_ ant-handler ids ([g rel-var] ...))
     (let ([g rel-var] ...)
       (lambda ids
         (ant-handler (g . ids) ...)))]
    [(_ ant-handler ids (let-pair ...) rel-exp0 rel-exp1 ...)
     (lift-ant-to-rel-aux ant-handler ids 
       (let-pair ... [g rel-exp0]) rel-exp1 ...)]))

;;;!
'(define-rel-lifted-comb extend-relation any)
(define-syntax extend-relation
  (syntax-rules ()
    [(_ ids . rel-exps)
      (lift-ant-to-rel-aux any ids () . rel-exps)]))]))

; The following  antecedent-to-relations 
; transformers are roughly equivalent. I don't know which is better.
; see examples below.

; (lift-to-relations ids (ant-comb rel rel ...))
(define-syntax lift-to-relations
  (syntax-rules ()
    [(_ ids (ant-comb rel ...))
     (lift-ant-to-rel-aux ant-comb ids () rel ...)]))

; (let-ants ids ((name rel) ...) body)
; NB: some macro systems do not like if 'ids' below is replaced by (id ...)
(define-syntax let-ants
  (syntax-rules ()
    [(_ ids ((ant-name rel-exp) ...) body)
     (lambda ids
       (let ((ant-name (rel-exp . ids)) ...)
         body))]))

; Unify lifted to be a binary relation
(define-syntax ==
  (syntax-rules (_)
    [(_ _ u)
     (lambda@ (sk) sk)]
    [(_ t _)
     (lambda@ (sk) sk)]
    [(_ t u)
     (lambda@ (sk fk subst)
       (let*-and (fk) ([subst (unify t u subst)])
         (@ sk fk subst)))]))



;------------------------------------------------------------------------
;;;;; Starts the real work of the system.

(define father  
  (relation ()
    (to-show 'jon 'sam)))

(pp (expand '(relation ()
    (to-show 'jon 'sam)))
)
(pp
(@ (father 'jon 'sam)
             initial-sk initial-fk empty-subst))

(test-check 'test-father0
  (let ([result
          (@ (father 'jon 'sam)
             initial-sk initial-fk empty-subst)])
    (and
      (equal? (car result) '())
      (equal? ((cdr result)) '())))
  #t)


(define child-of-male
  (relation (child dad)
    (to-show child dad)
    (father dad child)))

(test-check 'test-child-of-male-0
  (concretize-subst
    (car (@ (child-of-male 'sam 'jon)
	    initial-sk initial-fk empty-subst)))
  ;`(,(commitment 'child.0 'sam) ,(commitment 'dad.0 'jon)))
  '())  ; variables shouldn't leak


; The mark should be found here...
(define child-of-male-1
  (relation (child dad)
    (to-show child dad)
    (child-of-male dad child)))
(test-check 'test-child-of-male-1
  (concretize-subst
    (car (@ (child-of-male 'sam 'jon)
	    initial-sk initial-fk empty-subst)))
  ;`(,(commitment 'child.0 'sam) ,(commitment 'dad.0 'jon)))
  '())

(define rob/sal
  (relation ()
    (to-show 'rob 'sal)))

(define new-father
  (extend-relation (a1 a2) father rob/sal))

(test-check 'test-father-1
  (let ([result
	  (@ (new-father 'rob 'sal)
	    initial-sk initial-fk empty-subst)])
    (and
      (equal? (car result) '())
      (equal? ((cdr result)) '())))
  #t)


;	query (redo-k subst id ...) A SE ... -> result or '()
; The macro 'query' runs the antecedent A in the empty
; initial substitution, and reifies the resulting
; answer: the substitution and the redo-continuation bound
; to fresh variables with the names supplied by the user.
; The substitution and the redo continuation can then be used
; by Scheme expressions SE ...
; Before running the antecedent, the macro creates logical variables
; id ... for use in A and SE ...
; If the antecedent fails, '() is returned and SE ... are not evaluated.
; Note the similarity with shift/reset-based programming
; where the immediate return signifies "failure" and the invocation
; of the continuation a "success"
; Returning '() on failure makes it easy to create the list of answers

(define-syntax query
  (syntax-rules ()
    ((_ (redo-k subst id ...) A SE ...)
      (let-lv (id ...)
	(@ A
	  (lambda@ (redo-k subst) SE ...)
	  (lambda () '())
	  empty-subst)))))

(test-check 'test-father-2
  (query (redo-k subst x)
    (new-father 'rob x)
    (list (equal? (car subst) (commitment x 'sal)) (redo-k)))
  '(#t ()))


(test-check 'test-father-3
  (query (_ subst x)
    (new-father 'rob x)
    (concretize-subst/vars subst x))
  '((x.0 sal)))

(test-check 'test-father-4
  (query (_ subst x y)
    (new-father x y)
    (concretize-subst/vars subst x y))
  '((x.0 jon) (y.0 sam)))

(define rob/pat
  (relation ()
    (to-show 'rob 'pat)))

(define newer-father
  (extend-relation (a1 a2) new-father rob/pat))

(test-check 'test-father-5
  (query (redok subst x)
    (newer-father 'rob x)
    (pretty-print subst)
    (cons
      (concretize-subst/vars subst x)
      (redok)))
  '(((x.0 sal)) ((x.0 pat))))

(define stream-prefix
  (lambda (n strm)
    (if (null? strm) '()
      (cons (car strm)
        (if (zero? n) '()
          (stream-prefix (- n 1) ((cdr strm))))))))

; A simple version of query: useful as an illustration
; It returns a stream of concretized substitutions
(define-syntax query-stream
  (syntax-rules ()
    ((_ (id ...) A)
      (query (redo-k subst id ...) A (cons (concretize-subst subst) redo-k)))))

(test-check 'test-father-4-1
  (car (query-stream (x y)
	 (new-father x y)))
  '((x.0 . jon) (y.0 . sam) ))

(test-check 'test-father-5-1
  (let* ((ans1
	   (query-stream (x y)
	     (newer-father 'rob x)))
	  (ans2 ((cdr ans1)))
	  (ans3 ((cdr ans2))))
    (list (car ans1) (car ans2) ans3))
  '(((x.0 . sal)) ((x.0 . pat)) ()))


(define-syntax solve
  (syntax-rules ()
    [(_ n (var0 ...) ant)
      (if (<= n 0) '()
	(stream-prefix (- n 1)
	  (query (redo-k subst var0 ...)
	    ant
	    (cons (concretize-subst/vars subst var0 ...) redo-k))))]))

(define sam/rob
  (relation ()
    (to-show 'sam 'rob)))

(define newest-father (extend-relation (a1 a2) newer-father sam/rob))

(test-check 'test-father-6/solve
  (and
    (equal?
      (solve 5 (x) (newest-father 'rob x))
      '(((x.0 sal)) ((x.0 pat))))
    (equal?
      (solve 6 (x y) (newest-father x y))
      '(((x.0 jon) (y.0 sam))
        ((x.0 rob) (y.0 sal))
        ((x.0 rob) (y.0 pat))
        ((x.0 sam) (y.0 rob)))))
  #t)

; (define-syntax binary-intersect
;   (syntax-rules ()
;     ((_ ant1 ant2)
;       (lambda@ (sk)
; 	(@ ant1 (@ ant2 sk))))))

(define-syntax binary-intersect
  (syntax-rules ()
    [(_ ant1 ant2) (all ant1 ant2)]))

;;;!
'(define-rel-lifted-comb binary-intersect-relation binary-intersect)
(define-syntax binary-intersect-relation
  (syntax-rules ()
    [(_ ids . rel-exps)
      (lift-ant-to-rel-aux binary-intersect ids () . rel-exps)]))

(define parents-of-scouts
  (extend-relation (a1 a2)
    (fact () 'sam 'rob)
    (fact () 'roz 'sue)
    (fact () 'rob 'sal)))

(define parents-of-athletes
  (extend-relation (a1 a2)
    (fact () 'sam 'roz)
    (fact () 'roz 'sue)
    (fact () 'rob 'sal)))

(define busy-parents
  (binary-intersect-relation (a1 a2) 
    parents-of-scouts parents-of-athletes))

(define busy-parents
  (binary-intersect-relation (a1 a2) 
    parents-of-scouts parents-of-athletes))

(test-check 'test-busy-parents
  (solve 5 (x) (exists (y) (trace-ant-raw 'busy-parents (busy-parents x y))))
  '(((x.0 roz)) ((x.0 rob))))

; (define-syntax intersect-relation
;   (syntax-rules ()
;     [(_ (id ...) rel-exp) rel-exp]
;     [(_ (id ...) rel-exp0 rel-exp1 rel-exp2 ...)
;      (binary-intersect-relation (id ...) rel-exp0
;        (intersect-relation (id ...) rel-exp1 rel-exp2 ...))]))

;;;!
'(define-rel-lifted-comb intersect-relation all)
(define-syntax intersect-relation
  (syntax-rules ()
    [(_ ids . rel-exps)
      (lift-ant-to-rel-aux all ids () . rel-exps)]))

(define busy-parents
  (intersect-relation (a1 a2) parents-of-scouts parents-of-athletes))

(define conscientious-parents
  (extend-relation (a1 a2) parents-of-scouts parents-of-athletes))

(test-check 'test-conscientious-parents
  (solve 7 (x y) (conscientious-parents x y))
  '(((x.0 sam) (y.0 rob))
    ((x.0 roz) (y.0 sue))
    ((x.0 rob) (y.0 sal))
    ((x.0 sam) (y.0 roz))
    ((x.0 roz) (y.0 sue))
    ((x.0 rob) (y.0 sal))))

(define-syntax solution
  (syntax-rules ()
    [(_ (var0 ...) x)
     (let ([ls (solve 1 (var0 ...) x)])
       (if (null? ls) #f (car ls)))]))

(test-check 'test-father-7/solution
  (solution (x) (newest-father 'rob x))
  '((x.0 sal)))

(define grandpa-sam
  (relation (grandchild)
    (to-show grandchild)
    (exists (parent)
      (lambda (sk)
        ((newest-father 'sam parent)
         ((newest-father parent grandchild) sk))))))

(test-check 'test-grandpa-sam-1
  (solve 6 (y) (grandpa-sam y))
  '(((y.0 sal)) ((y.0 pat))))

(define grandpa-sam
  (relation ((once grandchild))
    (to-show grandchild)
    (exists (parent)
      (lambda (sk)
        ((newest-father 'sam parent)
         ((newest-father parent grandchild) sk))))))

(test-check 'test-grandpa-sam-1-1
  (solve 6 (y) (grandpa-sam y))
  '(((y.0 sal)) ((y.0 pat))))

(define child
  (relation ((once child) (once dad))
    (to-show child dad)
    (newest-father dad child)))

(test-check 'test-child-1
  (solve 10 (x y) (child x y))
  '(((x.0 sam) (y.0 jon))
    ((x.0 sal) (y.0 rob))
    ((x.0 pat) (y.0 rob))
    ((x.0 rob) (y.0 sam))))

(define grandpa-sam
  (relation (grandchild)
    (to-show grandchild)
    (exists (parent)
      (all
        (newest-father 'sam parent)
        (newest-father parent grandchild)))))

(test-check 'test-grandpa-sam-2
  (solve 6 (y) (grandpa-sam y))
  '(((y.0 sal)) ((y.0 pat))))

(define grandpa-maker
  (lambda (grandad)
    (relation ((once grandchild))
      (to-show grandchild)
      (exists (parent)
        (all
          (newest-father grandad parent)
          (newest-father parent grandchild))))))

(test-check 'test-grandpa-maker-1
  (solve 6 (x) ((grandpa-maker 'sam) x))
  '(((x.0 sal)) ((x.0 pat))))

(define grandpa-maker
  (lambda (guide* grandad*)
    (relation (grandchild)
      (to-show grandchild)
      (exists (parent)
        (all
          (guide* grandad* parent)
          (guide* parent grandchild))))))

(test-check 'test-grandpa-maker-2
  (solve 4 (x) ((grandpa-maker newest-father 'sam) x))
  '(((x.0 sal)) ((x.0 pat))))

(define grandpa
  (relation ((once grandad) (once grandchild))
    (to-show grandad grandchild)
    (exists (parent)
      (all
        (newest-father grandad parent)
        (newest-father parent grandchild)))))

(test-check 'test-grandpa-1
  (solve 4 (x) (grandpa 'sam x))
  '(((x.0 sal)) ((x.0 pat))))

(define father
  (extend-relation (a1 a2)
    (fact () 'jon 'sam)
    (extend-relation (a1 a2)
      (fact () 'sam 'rob)
      (extend-relation (a1 a2)
        (fact () 'sam 'roz)
        (extend-relation (a1 a2)
          (fact () 'rob 'sal)
          (fact () 'rob 'pat))))))

(define mother
  (extend-relation (a1 a2)
    (fact () 'roz 'sue)
    (fact () 'roz 'sid)))

(define grandpa
  (extend-relation (a1 a2)
    (relation (grandad grandchild)
      (to-show grandad grandchild)
      (exists (parent)
        (all
          (father grandad parent)
          (father parent grandchild))))
    (relation (grandad grandchild)
      (to-show grandad grandchild)
      (exists (parent)
        (all
          (father grandad parent)
          (mother parent grandchild))))))

(test-check 'test-grandpa-2
  (solve 10 (y) (grandpa 'sam y))
  '(((y.0 sal)) ((y.0 pat)) ((y.0 sue)) ((y.0 sid))))

(define grandpa/father
  (relation (grandad grandchild)
    (to-show grandad grandchild)
    (exists (parent)
      (all
        (father grandad parent)
        (father parent grandchild)))))

(define grandpa/mother
  (relation (grandad grandchild)
    (to-show grandad grandchild)
    (exists (parent)
      (all
        (father grandad parent)
        (mother parent grandchild)))))

(define grandpa
  (extend-relation (a1 a2) grandpa/father grandpa/mother))

(test-check 'test-grandpa-5
  (solve 10 (y) (grandpa 'sam y))
  '(((y.0 sal)) ((y.0 pat)) ((y.0 sue)) ((y.0 sid))))

(define grandpa-sam
  (let ([r (relation (child)
             (to-show child)
             (exists (parent)
               (all
                 (father 'sam parent)
                 (father parent child))))])
    (relation (child)
      (to-show child)
      (r child))))

(test-check 'test-grandpa-55
  (solve 6 (y) (grandpa-sam y))
  '(((y.0 sal)) ((y.0 pat))))

; The solution that used cuts
; (define grandpa/father
;   (relation/cut cut (grandad grandchild)
;     (to-show grandad grandchild)
;     (exists (parent)
;       (all
;         (father grandad parent)
;         (father parent grandchild)
;         cut))))
;
; (define grandpa/mother
;   (relation (grandad grandchild)
;     (to-show grandad grandchild)
;     (exists (parent)
;       (all
;         (father grandad parent)
;         (mother parent grandchild)))))


; Now we don't need it
(define grandpa/father
  (relation (grandad grandchild)
    (to-show grandad grandchild)
    (exists (parent)
      (all!
        (father grandad parent)
        (father parent grandchild)))))

(define grandpa/mother
  (relation (grandad grandchild)
    (to-show grandad grandchild)
    (exists (parent)
      (all
        (father grandad parent)
        (mother parent grandchild)))))

(define grandpa
  (lift-to-relations (a1 a2)
    (all!
      (extend-relation (a1 a2) grandpa/father grandpa/mother))))

(test-check 'test-grandpa-8
  (solve 10 (x y) (grandpa x y))
  '(((x.0 jon) (y.0 rob))))

; The solution that used to require cuts
; (define grandpa/father
;   (relation/cut cut (grandad grandchild)
;     (to-show grandad grandchild)
;     (exists (parent)
;       (all cut (father grandad parent) (father parent grandchild)))))

(define grandpa/father
  (relation (grandad grandchild)
    (to-show grandad grandchild)
    (exists (parent)
      (all (father grandad parent) (father parent grandchild)))))

; Properly, this requires soft cuts, aka *->, or Mercury's
; if-then-else. But we emulate it...
(define grandpa
  (let-ants (a1 a2) ([grandpa/father grandpa/father]
		     [grandpa/mother grandpa/mother])
    (if-only (succeeds grandpa/father) grandpa/father grandpa/mother)))

(test-check 'test-grandpa-10
  (solve 10 (x y) (grandpa x y))
  '(((x.0 jon) (y.0 rob))
    ((x.0 jon) (y.0 roz))
    ((x.0 sam) (y.0 sal))
    ((x.0 sam) (y.0 pat))))

(test-check 'test-grandpa-10-1
  (solve 10 (x) (grandpa x 'sue))
  '(((x.0 sam))))

; The same as above, with if-all! -- just to test the latter.
(define grandpa
  (let-ants (a1 a2) ([grandpa/father grandpa/father]
		     [grandpa/mother grandpa/mother])
    (if-only (all! (succeeds grandpa/father) (succeeds grandpa/father))
      grandpa/father grandpa/mother)))

(test-check 'test-grandpa-10
  (solve 10 (x y) (grandpa x y))
  '(((x.0 jon) (y.0 rob))
    ((x.0 jon) (y.0 roz))
    ((x.0 sam) (y.0 sal))
    ((x.0 sam) (y.0 pat))))

(test-check 'test-grandpa-10-1
  (solve 10 (x) (grandpa x 'sue))
  '(((x.0 sam))))

; Now do it with soft-cuts
(define grandpa
  (let-ants (a1 a2) ([grandpa/father grandpa/father]
		     [grandpa/mother grandpa/mother])
    (if-some grandpa/father succeed grandpa/mother)))

(test-check 'test-grandpa-10-soft-cut
  (solve 10 (x y) (grandpa x y))
  '(((x.0 jon) (y.0 rob))
    ((x.0 jon) (y.0 roz))
    ((x.0 sam) (y.0 sal))
    ((x.0 sam) (y.0 pat))))

(define a-grandma
  (relation (grandad grandchild)
    (to-show grandad grandchild)
    (exists (parent)
      (all! (mother grandad parent)))))

(define no-grandma-grandpa
  (let-ants (a1 a2) ([a-grandma a-grandma] [grandpa grandpa])
    (if-only a-grandma fail grandpa)))

(test-check 'test-no-grandma-grandpa-1
  (solve 10 (x) (no-grandma-grandpa 'roz x))
  '())

(define parents-of-scouts
  (extend-relation (a1 a2)
    (fact () 'sam 'rob)
    (fact () 'roz 'sue)
    (fact () 'rob 'sal)))

(define fathers-of-cubscouts
  (extend-relation (a1 a2)
    (fact () 'sam 'bob)
    (fact () 'tom 'adam)
    (fact () 'tad 'carl)))

(test-check 'test-partially-eval-sant
  (let-lv (p1 p2)
    (let* ([parents-of-scouts-sant
	     (ant->sant (parents-of-scouts p1 p2) empty-subst)]
           [cons@ (lambda@ (x y) (cons x y))]
           [split1 (@ 
                    partially-eval-sant parents-of-scouts-sant
                    cons@ (lambda () '()))]
           [a1 (car split1)]
           [split2 (@ partially-eval-sant (cdr split1) cons@
                     (lambda () '()))]
           [a2 (car split2)]
           [split3 (@ partially-eval-sant (cdr split2) cons@
                     (lambda () '()))]
           [a3 (car split3)])
      (map (lambda (subst)
             (concretize-subst/vars subst p1 p2))
	(list a1 a2 a3))))
  '(((p1.0 sam) (p2.0 rob)) ((p1.0 roz) (p2.0 sue)) ((p1.0 rob) (p2.0 sal))))

(define-syntax project
  (syntax-rules ()
    [(_ (var ...) ant)
     (lambda@ (sk fk subst)
       (let ([var (nonvar! (subst-in var subst))] ...)
	 (@ ant sk fk subst)))]))

(define-syntax project/no-check
  (syntax-rules ()
    [(_ (var ...) ant)
     (lambda@ (sk fk subst)
       (let ([var (subst-in var subst)] ...)
	 (@ ant sk fk subst)))]))

(define-syntax predicate
  (syntax-rules ()
    [(_ scheme-expression)
     (lambda@ (sk fk subst)
       (if scheme-expression (@ sk fk subst) (fk)))]))

(define nonvar!
  (lambda (t)
    (if (var? t)
      (errorf 'nonvar! "Logic variable ~s found after substituting."
	(concretize t))
      t)))

;   (syntax-rules ()
;     [(trace-vars title (var0 ...))
;      (promise-one-answer
;        (predicate/no-check (var0 ...)
;          (begin (display title) (display " ")
;                 (display '(var0 ...)) (display " ") (display (list var0 ...))
;                 (newline))))]))

(define-syntax trace-vars
  (syntax-rules ()
    [(_ title (var0 ...))
     (promise-one-answer
       (project/no-check (var0 ...)
         (predicate
	   (for-each 
	     (lambda (name val)
	       (cout title " " name ": " val nl))
             '(var0 ...) (concretize `(,var0 ...)))
	   )))]))

(define grandpa
  (relation (grandad grandchild)
    (to-show grandad grandchild)
    (exists (parent)
      (all 
        (father grandad parent)
        (project (parent)
          (all
            (predicate (starts-with-r? parent))
            (father parent grandchild)))))))

(define starts-with-r?
  (lambda (x)
    (and
      (symbol? x)
      (string=? (string (string-ref (symbol->string x) 0)) "r"))))

(test-check 'test-grandpa-11
  (solve 10 (x y) (grandpa x y))
  '(((x.0 sam) (y.0 sal)) ((x.0 sam) (y.0 pat))))

(define check
  (lambda (id f)
    (lambda term
      (if (not (procedure? f))
          (errorf id "Non-procedure found: ~s" f))
      (if (ormap var? term)
          (errorf id "Variable found: ~s" term))
      (apply f term))))

(test-check 'test-grandpa-12
  (solve 10 (x y) (grandpa x y))
  '(((x.0 sam) (y.0 sal)) ((x.0 sam) (y.0 pat))))

(define fun    
  (lambda (f)
    (fun-nocheck (check 'fun f))))

(define test1
  (lambda (x)
    (any (predicate (< 4 5))
      (== x (< 6 7)))))

(test-check 'test-test1
  (solution (x) (test1 x))
  '((x.0 x.0)))

(define test2
  (lambda (x)
    (any (predicate (< 5 4))
      (== x (< 6 7)))))

(test-check 'test-test2
  (solution (x) (test2 x))
  '((x.0 #t)))

(define test3
  (lambda (x y)
    (any
      (== x (< 5 4))
      (== y (< 6 7)))))

(test-check 'test-test3
  (solution (x y) (test3 x y))
  `((x.0 #f) (y.0 y.0)))

(define grandpa
  (relation (grandad grandchild)
    (to-show grandad grandchild)
    (exists (parent)
      (all
        (father grandad parent)
        (project (parent)
          (all
            (fails (predicate (starts-with-r? parent)))
            (father parent grandchild)))))))

(test-check 'test-grandpa-13
  (solve 10 (x y) (grandpa x y))
  '(((x.0 jon) (y.0 rob)) ((x.0 jon) (y.0 roz))))

(define instantiated
  (lambda (t)
    (project/no-check (t)
      (predicate (not (var? t))))))

(define view-subst
  (lambda (t)
    (lambda@ (sk fk subst)
      (pretty-print (subst-in t subst))
      (pretty-print (concretize-subst subst))
      (@ sk fk subst))))

(define grandpa
  (relation (grandad grandchild)
    (to-show grandad grandchild)
    (exists (parent)
      (all
        (father grandad parent)
        (father parent grandchild)
        (view-subst grandchild)))))

(test-check 'test-grandpa-14/view-subst
  (solve 10 (x y) (grandpa x y))
  (begin
    'rob
    '((grandad.0 x.0)
      (grandchild.0 y.0)
      (x.0 jon)
      (parent.0 sam)
      (y.0 rob))
    'roz
    '((grandad.0 x.0)
      (grandchild.0 y.0)
      (x.0 jon)
      (parent.0 sam)
      (y.0 roz))
    'sal 
    '((grandad.0 x.0)
      (grandchild.0 y.0)
      (x.0 sam)
      (parent.0 rob)
      (y.0 sal))
    'pat
    '((grandad.0 x.0)
      (grandchild.0 y.0)
      (x.0 sam)
      (parent.0 rob)
      (y.0 pat))
    '(((x.0 jon) (y.0 rob))
      ((x.0 jon) (y.0 roz))
      ((x.0 sam) (y.0 sal))
      ((x.0 sam) (y.0 pat)))))

(define father
  (extend-relation (a1 a2) father
    (extend-relation (a1 a2) (fact () 'jon 'hal)
      (extend-relation (a1 a2) (fact () 'hal 'ted) (fact () 'sam 'jay)))))

(define ancestor
  (extend-relation (a1 a2)
    (relation (old young)
      (to-show old young)
      (father old young))
    (relation (old young)
      (to-show old young)
      (exists (not-so-old)
        (all (father old not-so-old) (ancestor not-so-old young))))))

(test-check 'test-ancestor
  (solve 21 (x) (ancestor 'jon x))
  '(((x.0 sam))
    ((x.0 hal))
    ((x.0 rob))
    ((x.0 roz))
    ((x.0 jay))
    ((x.0 sal))
    ((x.0 pat))
    ((x.0 ted))))

(define common-ancestor
  (relation (young-a young-b old)
    (to-show young-a young-b old)
    (all
      (ancestor old young-a)
      (ancestor old young-b))))

(test-check 'test-common-ancestor
  (solve 4 (x) (common-ancestor 'pat 'jay x))
  '(((x.0 jon)) ((x.0 sam))))

(define younger-common-ancestor
  (relation (young-a young-b old not-so-old)
    (to-show young-a young-b old not-so-old)
    (all
      (common-ancestor young-a young-b not-so-old)
      (common-ancestor young-a young-b old)
      (ancestor old not-so-old))))

(test-check 'test-younger-common-ancestor
  (solve 4 (x) (younger-common-ancestor 'pat 'jay 'jon x))
  '(((x.0 sam))))

(define youngest-common-ancestor
  (relation (young-a young-b not-so-old)
    (to-show young-a young-b not-so-old)
    (all
      (common-ancestor young-a young-b not-so-old)
      (exists (y)
        (fails (younger-common-ancestor young-a young-b not-so-old y))))))

(test-check 'test-youngest-common-ancestor
  (solve 4 (x) (youngest-common-ancestor 'pat 'jay x))
  '(((x.0 sam))))

(test-check 'test-Seres-Spivey
  (let ([father
	  (lambda (dad child)
	    (any
	      (all (== dad 'jon) (== child 'sam))
	      (all (== dad 'sam) (== child 'rob))
	      (all (== dad 'sam) (== child 'roz))
	      (all (== dad 'rob) (== child 'sal))
	      (all (== dad 'rob) (== child 'pat))
	      (all (== dad 'jon) (== child 'hal))
	      (all (== dad 'hal) (== child 'ted))
	      (all (== dad 'sam) (== child 'jay))))])
    (letrec
        ([ancestor
           (lambda (old young)
             (any
               (father old young)
               (exists (not-so-old)
                 (all
                   (father old not-so-old)
                   (ancestor not-so-old young)))))])
      (solve 20 (x) (ancestor 'jon x))))
  '(((x.0 sam))
    ((x.0 hal))
    ((x.0 rob))
    ((x.0 roz))
    ((x.0 jay))
    ((x.0 sal))
    ((x.0 pat))
    ((x.0 ted))))

(define towers-of-hanoi
  (letrec
      ([move
         (extend-relation (a1 a2 a3 a4)
           (fact () 0 _ _ _)
           (relation (n a b c)
             (to-show n a b c)
             (project (n)
               (if-only (predicate (positive? n))
                 (let ([m (- n 1)])
                   (all 
                     (move m a c b)
                     (project (a b)
                       (begin
                         (cout "Move a disk from " a " to " b nl)
                         (move m c b a)))))))))])
    (relation (n)
      (to-show n)
      (move n 'left 'middle 'right))))

(cout "test-towers-of-hanoi with 3 disks: "
  (solution () (towers-of-hanoi 3))
  nl nl
  )

'(test-check 'test-compose-subst-5
  (let-lv (x y z)
    (equal?
      (let ([term `(p ,x ,y (g ,z))])
	(let ([s (compose-subst (unit-subst y z) (unit-subst x `(f ,y)))]
	      [r (compose-subst (unit-subst x 'a) (unit-subst z 'b))])
	  (let ([term1 (subst-in term s)])
	    (write term1)
	    (newline)
	    (let ([term2 (subst-in term1 r)])
	      (write term2)
	      (newline)
	      (let ([sr (compose-subst s r)])
		(write sr)
		(newline)
		(subst-in term sr))))))
      (begin
	`(p (f ,y) ,z (g ,z))
	`(p (f ,y) b (g b))
	`(,(commitment y 'b) ,(commitment x `(f ,y)) ,(commitment z 'b))
	`(p (f ,y) b (g b)))))
  #t)

(test-check 'test-unify/pairs
  (let-lv (w x y z u)
    (and
      (and
        (equal?
          (unify `(,x ,4) `(3 ,x) empty-subst)
          #f)
        (equal?
          (unify `(,x ,x) '(3 4) empty-subst)
          #f))
;       (and
;         (equal?
;           (unify `(,x ,y) '(3 4) empty-subst)
;           `(,(commitment x 3) ,(commitment y 4)))
;         (equal?
;           (unify `(,x 4) `(3 ,y) empty-subst)
;          `(,(commitment x 3) ,(commitment y 4))))
;       (and
;         (equal?
;           (unify `(,x 4 3 ,w) `(3 ,y ,x ,z) empty-subst)
;           `(,(commitment x 3) ,(commitment y 4) ,(commitment w z)))
;         (equal?
;           (unify `(,x 4) `(,y ,y) empty-subst)
;           `(,(commitment x 4) ,(commitment y 4)))
;         (equal?
;           (unify `(,x 4 3) `(,y ,y ,x) empty-subst)
;           #f)
;         (equal?
;           (unify `(,w (,x (,y ,z) 8)) `(,w (,u (abc ,u) ,z)) empty-subst)
;           `(,(commitment x 8)
;             ,(commitment y 'abc)
;             ,(commitment z 8)
;             ,(commitment u 8)))
;         (equal?
;           (unify `(p (f a) (g ,x)) `(p ,x ,y) empty-subst)
;           `(,(commitment x '(f a)) ,(commitment y '(g (f a)))))
;         (equal?
;           (unify `(p (g ,x) (f a)) `(p ,y ,x) empty-subst)
;           `(,(commitment y '(g (f a))) ,(commitment x '(f a))))
;         (equal?
;           (unify `(p a ,x (h (g ,z))) `(p ,z (h ,y) (h ,y)) empty-subst)
;           `(,(commitment z 'a)
;             ,(commitment x '(h (g a)))
;             ,(commitment y '(g a))))
;         (equal? ;;; Oleg's succeeds on this.
;           (unify `(p ,x ,x) `(p ,y (f ,y)) empty-subst)
;           #f)))
      ))
  #t)

(define towers-of-hanoi-path
  (let ([steps '()])
    (let ([push-step (lambda (x y) (set! steps (cons `(,x ,y) steps)))])
      (letrec
          ([move
             (extend-relation (a1 a2 a3 a4)
               (fact () 0 _ _ _)
               (relation (n a b c)
                 (to-show n a b c)
                 (project (n)
                   (if-only (predicate (positive? n))
                     (let ([m (- n 1)])
                       (all
                         (move m a c b)
                         (project (a b)
                           (begin
                             (push-step a b)
                             (move m c b a)))))))))])
        (relation (n path)
          (to-show n path)
          (begin
            (set! steps '())
            (any
              (fails (move n 'l 'm 'r))
              (== path (reverse steps)))))))))

(test-check 'test-towers-of-hanoi-path
  (solution (path) (towers-of-hanoi-path 3 path))
  '((path.0 ((l m) (l r) (m r) (l m) (r l) (r m) (l m)))))


;------------------------------------------------------------------------
(test-check 'test-unify/pairs-oleg1
  (let-lv (x y)
    (unify `(,x ,4) `(3 ,x) empty-subst))
  #f)

(test-check 'test-unify/pairs-oleg2
  (let-lv (x y)
    (unify `(,x ,x) '(3 4) empty-subst))
  #f)

(let-lv (x y)
  (test-check 'test-unify/pairs-oleg3
    (unify `(,x ,y) '(3 4) empty-subst)
    `(,(commitment y 4) ,(commitment x 3))))

(let-lv (x y)
  (test-check 'test-unify/pairs-oleg4
    (unify `(,x 4) `(3 ,y) empty-subst)
    `(,(commitment y 4) ,(commitment x 3))))

(let-lv (x y w z)
  (test-check 'test-unify/pairs-oleg5
    (let ([s (normalize-subst
	       (unify `(,x 4 3 ,w) `(3 ,y ,x ,z) empty-subst))])
      (let ([vars (list w y x)])
	(map commitment
	  vars
	  (subst-vars-recursively vars s))))
    `(,(commitment w z)
       ,(commitment y 4)
       ,(commitment x 3))))

(let-lv (x y w z)
  (test-check 'test-unify/pairs-oleg6
    (let ([s (normalize-subst
	       (unify `(,x 4) `(,y ,y) empty-subst))])
      (let ([vars (list y x)])
	(map commitment
	  vars
	  (subst-vars-recursively vars s))))
    `(,(commitment y 4) ,(commitment x 4))))

(test-check 'test-unify/pairs-oleg7
  (let-lv (x y)
    (unify `(,x 4 3) `(,y ,y ,x) empty-subst))
    #f)

(let-lv (x y w z u)
  (test-check 'test-unify/pairs-oleg8
    (let ([s (normalize-subst
	       (unify
		 `(,w (,x (,y ,z) 8))
		 `(,w (,u (abc ,u) ,z))
		 empty-subst))])
      (let ([vars (list u z y x)])
	(map commitment
	  vars
	  (subst-vars-recursively vars s))))
    `(,(commitment u 8)
       ,(commitment z 8)
       ,(commitment y 'abc)
       ,(commitment x 8))))

(let-lv (x y w z u)
  (test-check 'test-unify/pairs-oleg8
    (let ([s (normalize-subst
	       (unify `(p (f a) (g ,x)) `(p ,x ,y) empty-subst))])
      (let ([vars (list y x)])
	(map commitment
	  vars
	  (subst-vars-recursively vars s))))
    `(,(commitment y '(g (f a)))
       ,(commitment x '(f a)))))

(let-lv (x y w z u)
  (test-check 'test-unify/pairs-oleg10
    (let ([s (normalize-subst
	       (unify `(p (g ,x) (f a)) `(p ,y ,x) empty-subst))])
      (let ([vars (list x y)])
	(map commitment
	  vars
	  (subst-vars-recursively vars s))))
    `(,(commitment x '(f a))
       ,(commitment y '(g (f a))))))

(let-lv (x y w z u)
  (test-check 'test-unify/pairs-oleg11
    (let ([s (normalize-subst
	       (unify
		 `(p a ,x (h (g ,z)))
		 `(p ,z (h ,y) (h ,y))
		 empty-subst))])
      (let ([vars (list y x z)])
	(map commitment
	  vars
	  (subst-vars-recursively vars s))))
    `(,(commitment y '(g a))
       ,(commitment x '(h (g a)))
       ,(commitment z 'a))))

; The following loops...
; (concretize-subst
;   (let-lv (x y)
;     (let* ((s (unify x `(1 ,x) '()))
; 	   (s (unify y `(1 ,y) s))
; 	   (s (unify x y s))) s)))



; (let-lv (x y w z u)
;   (test-check 'test-unify/pairs-oleg12
;     (concretize-subst ;;; was #f
;       (let ([s (unify `(p ,x ,x) `(p ,y (f ,y)) empty-subst)])
; 	(let ([var (map commitment->var s)])
; 	  (map commitment
; 	    var
; 	    (subst-vars-recursively var s)))))
;     `(;,(commitment '*d.0 '())
;        ,(commitment '*a.0 '(f *a.0))
;        ;,(commitment '*d.1 '((f . *d.1)))
;        ,(commitment '*d.0 '((f . *d.0)))
;        ;,(commitment '*a.1 'f)
;        ;,(commitment 'y.0  '(f (f . *d.1)))
;        ,(commitment 'y.0  '(f (f . *d.0)))
;        ,(commitment 'x.0  '(f (f . *d.0))))))

; (let-lv (x y w z u)
;   (test-check 'test-unify/pairs-oleg13
;     (concretize-subst ;;; was #f
;       (let ([s (unify `(p ,x ,x) `(p ,y (f ,y)) empty-subst)])
; 	(let ([var (map commitment->var s)])
; 	  (map commitment
; 	    var
; 	    (subst-vars-recursively var s)))))
;     `(;,(commitment '*d.0 '())
;        ,(commitment '*a.0 '(f *a.0))
;        ;,(commitment '*d.1 '((f . *d.1)))
;        ,(commitment '*d.0 '((f . *d.0)))
;        ;,(commitment '*a.1 'f)
;        ;,(commitment 'y.0  '(f (f . *d.1)))
;        ,(commitment 'y.0  '(f (f . *d.0)))
;        ,(commitment 'x.0  '(f (f . *d.0))))))

(test-check 'test-fun-resubst-oleg
  (concretize
    (let ([j (relation (x w z)
	       (to-show z)
               (let ([x 4]
                     [w 3])
                 (== z (cons x w))))])
      (solve 4 (q) (j q))))
  '(((q.0 (4 . 3)))))
          
;Baader & Snyder
(test-check 'test-pathological-oleg
  (list
    (let-lv (x0 x1 y0 y1)
      (pretty-print
	(concretize-subst
	  (let ([s (unify
		     `(h ,x1 (f ,y0 ,y0) ,y1)
		     `(h (f ,x0 ,x0) ,y1 ,x1)
		     empty-subst)])
	    (let ([vars (map commitment->var s)])
	      (map commitment
		vars
		(subst-vars-recursively vars s))))))
      (newline) #t)

    (let-lv (x0 x1 x2 y0 y1 y2)
      (pretty-print
	(concretize-subst
	  (let ([s (unify
		     `(h ,x1 ,x2 (f ,y0 ,y0) (f ,y1 ,y1) ,y2)
		     `(h (f ,x0 ,x0) (f ,x1 ,x1) ,y1 ,y2 ,x2)
		     empty-subst)])
	    (let ([vars (map commitment->var s)])
	      (map commitment
		vars
		(subst-vars-recursively vars s))))))
      (newline) #t)

    (let-lv (x0 x1 x2 x3 x4 y0 y1 y2 y3 y4)
      (pretty-print
	(concretize-subst
	  (let ([s (unify
		     `(h ,x1 ,x2 ,x3 ,x4 (f ,y0 ,y0) (f ,y1 ,y1)
			(f ,y2 ,y2) (f ,y3 ,y3) ,y4)
		     `(h (f ,x0 ,x0) (f ,x1 ,x1) (f ,x2 ,x2)
			(f ,x3 ,x3) ,y1 ,y2 ,y3 ,y4 ,x4)
		     empty-subst)])
	    (let ([vars (map commitment->var s)])
	      (map commitment
		vars
		(subst-vars-recursively vars s))))))
      #t))
  (list #t #t #t))

(test-check 'unification-of-free-vars-1
  (solve 1 (x)
    (let-lv (y)
      (all!! (== x y) (== y 5))))
  '(((x.0 5))))

(test-check 'unification-of-free-vars-2
  (solve 1 (x)
    (let-lv (y)
      (all!! (== y 5) (== x y))))
  '(((x.0 5))))

(test-check 'unification-of-free-vars-3
  (solve 1 (x)
    (let-lv (y)
      (all!! (== y x) (== y 5))))
  '(((x.0 5))))

(test-check 'unification-of-free-vars-3
  (solve 1 (x)
    (let-lv (y)
      (all!! (== x y) (== y 5) (== x y))))
  '(((x.0 5))))

(test-check 'unification-of-free-vars-4
  (solve 1 (x)
    (exists (y)
      (all! (== y x) (== y 5) (== x y))))
  '(((x.0 5))))

(test-check 'length-of-subst
  (let-lv (x y z)
    (let* ((subst (unify x `(1 2 3 4 5 ,z) '()))
	    (subst (unify x `(1 . ,y) subst))
	    (subst (unify z 42 subst)))
      (concretize-subst subst)))
  '((z.0 . 42) (y.0 2 3 4 5 z.0) (x.0 1 2 3 4 5 z.0)))
  ;'((z.0 . 42) (y.0 2 3 4 5 a*.0) (a*.0 . z.0) (x.0 1 2 3 4 5 a*.0)))

(define concat
  (lambda (xs ys)
    (cond
      [(null? xs) ys]
      [else (cons (car xs) (concat (cdr xs) ys))])))

(test-check 'test-concat-as-function
  (concat '(a b c) '(u v))
  '(a b c u v))

(test-check 'test-fun-concat
  (solve 1 (q)
    (== q (concat '(a b c) '(u v))))
  '(((q.0 (a b c u v)))))

(define concat
  (extend-relation (a1 a2 a3)
    (fact (xs) '() xs xs)
    (relation (x xs (once ys) zs)
      (to-show `(,x . ,xs) ys `(,x . ,zs))
      (concat xs ys zs))))

(test-check 'test-concat
  (time
    (and
      (equal?
        (solve 6 (q) (concat '(a b c) '(u v) q))
        '(((q.0 (a b c u v)))))
      (equal?
        (solve 6 (q) (concat '(a b c) q '(a b c u v)))
        '(((q.0 (u v)))))
      (equal?
        (solve 6 (q) (concat q '(u v) '(a b c u v)))
        '(((q.0 (a b c)))))
      (equal?
        (solve 6 (q r) (concat q r '(a b c u v)))
        '(((q.0 ()) (r.0 (a b c u v)))
          ((q.0 (a)) (r.0 (b c u v)))
          ((q.0 (a b)) (r.0 (c u v)))
          ((q.0 (a b c)) (r.0 (u v)))
          ((q.0 (a b c u)) (r.0 (v)))
          ((q.0 (a b c u v)) (r.0 ()))))
      (equal?
        (solve 6 (q r s) (concat q r s))
        '(((q.0 ()) (r.0 r.0) (s.0 r.0))
          ((q.0 (x.0)) (r.0 r.0) (s.0 (x.0 . r.0)))
          ((q.0 (x.0 x.1)) (r.0 r.0) (s.0 (x.0 x.1 . r.0)))
          ((q.0 (x.0 x.1 x.2)) (r.0 r.0) (s.0 (x.0 x.1 x.2 . r.0)))
          ((q.0 (x.0 x.1 x.2 x.3))
           (r.0 r.0)
           (s.0 (x.0 x.1 x.2 x.3 . r.0)))
          ((q.0 (x.0 x.1 x.2 x.3 x.4))
           (r.0 r.0)
           (s.0 (x.0 x.1 x.2 x.3 x.4 . r.0)))))
      (equal?
        (solve 6 (q r) (concat q '(u v) `(a b c . ,r)))
        '(((q.0 (a b c)) (r.0 (u v)))
          ((q.0 (a b c x.0)) (r.0 (x.0 u v)))
          ((q.0 (a b c x.0 x.1)) (r.0 (x.0 x.1 u v)))
          ((q.0 (a b c x.0 x.1 x.2)) (r.0 (x.0 x.1 x.2 u v)))
          ((q.0 (a b c x.0 x.1 x.2 x.3)) (r.0 (x.0 x.1 x.2 x.3 u v)))
          ((q.0 (a b c x.0 x.1 x.2 x.3 x.4))
           (r.0 (x.0 x.1 x.2 x.3 x.4 u v)))))
      (equal?
        (solve 6 (q) (concat q '() q))
        '(((q.0 ()))
          ((q.0 (x.0)))
          ((q.0 (x.0 x.1)))
          ((q.0 (x.0 x.1 x.2)))
          ((q.0 (x.0 x.1 x.2 x.3)))
          ((q.0 (x.0 x.1 x.2 x.3 x.4)))))))
  #t)

; Using the properties of the unifier to do the proper garbage
; collection of logical vars

(test-check 'lv-elim-1
  (concretize
    (let-lv (x z dummy)
      (@ 
	(exists (y)
	  (== `(,x ,z ,y) `(5 9 ,x)))
	(lambda@ (fk subst) subst)
	initial-fk
	(unit-subst dummy 'dummy))))
  '((y.0 . 5) (z.0 . 9) (x.0 . 5) (dummy.0 . dummy)))
  ;'((z.0 . 9) (x.0 . 5) (dummy.0 . dummy)))

(test-check 'lv-elim-2
  (concretize
    (let-lv (x dummy)
      (@ 
	(exists (y)
	  (== `(,x ,y) `((5 ,y) ,7)))
	(lambda@ (fk subst) subst)
	initial-fk
	(unit-subst dummy 'dummy))))
  '((y.0 . 7) (x.0 5 y.0) (dummy.0 . dummy)))
  ;'((a*.0 . 7) (x.0 5 a*.0) (dummy.0 . dummy)))

; verifying corollary 2 of proposition 10
(test-check 'lv-elim-3
  (concretize
    (let-lv (x v dummy)
      (@ 
	(exists (y)
	  (== x `(a b c ,v d)))
	(lambda@ (fk subst) subst)
	initial-fk
	(unit-subst dummy 'dummy))))
  '((x.0 a b c v.0 d) (dummy.0 . dummy)))
  ;'((a*.0 . v.0) (x.0 a b c a*.0 d) (dummy.0 . dummy)))

; pruning several variables sequentially and in parallel
(test-check 'lv-elim-4-1
  (concretize
    (let-lv (x v b dummy)
      (@ 
	(let-lv (y)
	  (== `(,b ,x ,y) `(,x ,y 1)))
	(lambda@ (fk subst) subst)
	initial-fk
	(unit-subst dummy 'dummy))))
  '((y.0 . 1) (x.0 . y.0) (b.0 . x.0) (dummy.0 . dummy)))

; (test-check 'lv-elim-4-2
;   (concretize
;     (let-lv (v b dummy)
;       (@ 
; 	(exists (x)
; 	  (exists (y)
; 	    (== `(,b ,x ,y) `(,x ,y 1))))
; 	  (lambda@ (fk subst) subst)
; 	  initial-fk
; 	  (unit-subst dummy 'dummy))))
;     '((b.0 . 1) (dummy.0 . dummy)))

; (test-check 'lv-elim-4-3
;   (concretize
;     (let-lv (v b dummy)
;       (@ 
; 	(exists (y)
; 	  (exists (x)
; 	    (== `(,b ,x ,y) `(,x ,y 1))))
; 	  (lambda@ (fk subst) subst)
; 	  initial-fk
; 	  (unit-subst dummy 'dummy))))
;     '((b.0 . 1) (dummy.0 . dummy)))

(test-check 'lv-elim-4-4
  (concretize
    (let-lv (v b dummy)
      (@ 
	(exists (x y)
	    (== `(,b ,x ,y) `(,x ,y 1)))
	  (lambda@ (fk subst) subst)
	  initial-fk
	  (unit-subst dummy 'dummy))))
  '((y.0 . 1) (x.0 . y.0) (b.0 . x.0) (dummy.0 . dummy)))
  ;'((b.0 . 1) (dummy.0 . dummy)))

; pruning several variables sequentially and in parallel
; for indirect (cyclic) dependency
(test-check 'lv-elim-5-1
  (concretize
    (let-lv (x v b dummy)
      (@ 
	(let-lv (y)
	  (== `(,b ,y ,x) `(,x (1 ,x) ,y)))
	(lambda@ (fk subst) subst)
	initial-fk
	(unit-subst dummy 'dummy))))
  '((x.0 1 x.0) (y.0 1 x.0) (b.0 . x.0) (dummy.0 . dummy)))
  ;'((x.0 1 a*.0) (a*.0 . x.0) (y.0 1 a*.0) (b.0 . x.0) (dummy.0 . dummy)))

; (test-check 'lv-elim-5-2
;   (concretize
;     (let-lv (v b dummy)
;       (@ 
; 	(exists (x)
; 	  (exists (y)
; 	  (== `(,b ,y ,x) `(,x (1 ,x) ,y))))
; 	(lambda@ (fk subst) subst)
; 	initial-fk
; 	(unit-subst dummy 'dummy))))
;   '((a*.0 1 a*.0) (b.0 1 a*.0) (dummy.0 . dummy)))

; (test-check 'lv-elim-5-3
;   (concretize
;     (let-lv (v b dummy)
;       (@ 
; 	(exists (y)
; 	  (exists (x)
; 	  (== `(,b ,y ,x) `(,x (1 ,x) ,y))))
; 	(lambda@ (fk subst) subst)
; 	initial-fk
; 	(unit-subst dummy 'dummy))))
;   '((a*.0 1 a*.0) (b.0 1 a*.0) (dummy.0 . dummy)))

(test-check 'lv-elim-5-4
  (concretize
    (let-lv (v b dummy)
      (@ 
	(exists (x y)
	  (== `(,b ,y ,x) `(,x (1 ,x) ,y)))
	(lambda@ (fk subst) subst)
	initial-fk
	(unit-subst dummy 'dummy))))
  '((x.0 1 x.0) (y.0 1 x.0) (b.0 . x.0) (dummy.0 . dummy)))
 ;'((a*.0 1 a*.0) (b.0 1 a*.0) (dummy.0 . dummy)))

; We should only be concerned about a direct dependency:
;  ((x . y) (y . (1 t)) (t . x) (a . x))
; pruning x and y in sequence or in parallel gives the same result:
;  ((t . (1 t)) (a . (1 t)))


;;;; *****************************************************************
;;;; This is the start of a different perspective on logic programming.
;;;; Unitl I saw (!! fk), it was not possible to do the things that I
;;;; wanted to do.  This does obviate the need for unify-sequence, so
;;;; that is good.

(define father
  (lambda (dad child)
    (all!!
      (== dad 'jon)
      (== child 'sam))))

(define father
  (extend-relation (a1 a2) father
    (lambda (dad child)
      (all!! 
	(== dad 'sam)
	(== child 'rob)))))

(define father
  (extend-relation (a1 a2) father
    (lambda (dad child)
      (all!!
	(== dad 'rob)
	(== child 'sal)))))

(define grandpa
  (lambda (grandfather child)
    (exists (x)
      (all (father grandfather x) (father x child)))))

(test-check 'grandpa-ng
  (solve 5 (x y) (grandpa x y))
  '(((x.0 jon) (y.0 rob)) ((x.0 sam) (y.0 sal))))

(define father-rob-sal (fact () 'rob 'sal))
(define father-sam-rob (fact () 'sam 'rob))
(define father-jon-sam (fact () 'jon 'sam))
(define father
  (extend-relation (a1 a2) father-jon-sam father-sam-rob father-rob-sal))

(test-check 'grandpa-ng-1
  (solve 5 (x y) (grandpa x y))
  '(((x.0 jon) (y.0 rob)) ((x.0 sam) (y.0 sal))))

(test-check 'grandpa-ng-2
  (solve 5 (y) (grandpa 'sam y))
  '(((y.0 sal))))

(test-check 'grandpa-ng-2
  (solve 5 (x) (grandpa x 'sal))
  '(((x.0 sam))))

; No need for the following any more...
; (define grandpa-sam
;   (lambda (child)
;     (lambda@ (sk fk subst)
;       (let ([next (!! fk)]
;             [cut (!! cutk)])
;         (let-lv (grandfather)
;           (@ (all
;                (all next (== grandfather 'sam))
;                cut
;                (exists (x)
;                  (all (father grandfather x) (father x child))))
;              sk fk subst ))))))

(define grandpa-sam
  (lambda (child)
    (exists (grandfather)
      (if-only (== grandfather 'sam)
	(exists (x)
	  (all (father grandfather x) (father x child)))))))

(test-check 'grandpa-sam-1
  (solve 5 () (grandpa-sam 'sal))
  '(()))

(define grandpa-sam
  (relation (child)
    (to-show child)
    (exists (x)
      (all (father 'sam x) (father x child)))))

(test-check 'grandpa-sam-2
  (solve 5 () (grandpa-sam 'sal))
  '(()))

(define father-rob-sal (fact () 'rob 'sal))
(define father-sam-rob (fact () 'sam 'rob))
(define father-jon-sam (fact () 'jon 'sam))
(define father 
  (extend-relation (a1 a2) father-jon-sam father-sam-rob father-rob-sal))

(test-check 'grandpa-sam-3
  (solve 5 () (grandpa-sam 'sal))
  '(()))

(define father-rob-sal
  (relation () (to-show 'rob 'sal)))

(define father-sam-rob
  (relation () (to-show 'sam 'rob)))

(define father-johh-sam
  (relation () (to-show 'jon 'sam)))

(define father 
  (extend-relation (a1 a2) father-jon-sam father-sam-rob father-rob-sal))

(test-check 'grandpa-sam-4
  (solve 5 () (grandpa-sam 'sal))
  '(()))

;;;; This verifies that the idea works.  Now, to run the book through this
;;;; system.

; Extending relations in truly mathematical sense.
; First, why do we need this.

(define fact1 (fact () 'x1 'y1))
(define fact2 (fact () 'x2 'y2))
(define fact3 (fact () 'x3 'y3))
(define fact4 (fact () 'x4 'y4))

; R1 and R2 are overlapping
(define R1 (extend-relation (a1 a2) fact1 fact2))
(define R2 (extend-relation (a1 a2) fact1 fact3)) 

(cout nl "R1:" nl)
(pretty-print (solve 10 (x y) (R1 x y)))
(cout nl "R2:" nl)
(pretty-print (solve 10 (x y) (R2 x y)))
(cout nl "R1+R2:" nl)
(pretty-print 
  (solve 10 (x y)
    ((extend-relation (a1 a2) R1 R2) x y)))

; Infinitary relation
; r(z,z).
; r(s(X),s(Y)) :- r(X,Y).

(define Rinf
  (extend-relation (a1 a2)
    (fact () 'z 'z)
    (relation (x y t1 t2)
      (to-show t1 t2)
      (all
	(== t1 `(s ,x))
	(== t2 `(s ,y))
	(Rinf x y)))))
(cout nl "Rinf:" nl)
(time (pretty-print (solve 5 (x y) (Rinf x y))))
(cout nl "Rinf+R1: Rinf starves R1:" nl)
(time
  (pretty-print 
    (solve 5 (x y)
      ((extend-relation (a1 a2) Rinf R1) x y))))

; Solving starvation problem: extend R1 and R2 so that they
; are interleaved
; ((sf-extend R1 R2) sk fk)
; (R1 sk fk)
; If R1 fails, we try the rest of R2
; If R1 succeeds, it executes (sk fk)
; with fk to re-prove R1. Thus fk is the "rest" of R1
; So we pass sk (lambda () (run-rest-of-r2 interleave-with-rest-of-r1))
; There is a fixpoint in the following algorithm!
; Or a second-level shift/reset!

(test-check "Rinf+R1"
  (time 
    (solve 7 (x y)
      (any-interleave (Rinf x y) (R1 x y))))
  '(((x.0 z) (y.0 z))
     ((x.0 x1) (y.0 y1))
     ((x.0 (s z)) (y.0 (s z)))
     ((x.0 x2) (y.0 y2))
     ((x.0 (s (s z))) (y.0 (s (s z))))
     ((x.0 (s (s (s z)))) (y.0 (s (s (s z)))))
     ((x.0 (s (s (s (s z))))) (y.0 (s (s (s (s z)))))))
  )

(test-check "R1+Rinf"
  (time 
    (solve 7 (x y)
      (any-interleave (R1 x y) (Rinf x y))))
  '(((x.0 x1) (y.0 y1))
    ((x.0 z) (y.0 z))
    ((x.0 x2) (y.0 y2))
    ((x.0 (s z)) (y.0 (s z)))
    ((x.0 (s (s z))) (y.0 (s (s z))))
    ((x.0 (s (s (s z)))) (y.0 (s (s (s z)))))
    ((x.0 (s (s (s (s z))))) (y.0 (s (s (s (s z)))))))
)


(test-check "R2+R1"
  (solve 7 (x y)
    (any-interleave (R2 x y) (R1 x y)))
  '(((x.0 x1) (y.0 y1))
    ((x.0 x1) (y.0 y1))
    ((x.0 x3) (y.0 y3))
    ((x.0 x2) (y.0 y2)))
)

(test-check "R1+fact3"
  (solve 7 (x y)
    (any-interleave (R1 x y) (fact3 x y)))
    '(((x.0 x1) (y.0 y1)) ((x.0 x3) (y.0 y3)) ((x.0 x2) (y.0 y2)))
)

(test-check "fact3+R1"
  (solve 7 (x y)
    (any-interleave (fact3 x y) (R1 x y)))
    '(((x.0 x3) (y.0 y3)) ((x.0 x1) (y.0 y1)) ((x.0 x2) (y.0 y2)))
)

;;; Test for nonoverlapping.

(cout nl "any-union" nl)
(test-check "R1+R2"
  (solve 10 (x y)
    (any-union (R1 x y) (R2 x y)))
  '(((x.0 x1) (y.0 y1))
    ((x.0 x2) (y.0 y2))
    ((x.0 x3) (y.0 y3))))

(test-check "R2+R1"
  (solve 10 (x y)
    (any-union (R2 x y) (R1 x y)))
  '(((x.0 x1) (y.0 y1))
    ((x.0 x3) (y.0 y3))
    ((x.0 x2) (y.0 y2))))

(test-check "R1+R1"
  (solve 10 (x y)
    (any-union (R1 x y) (R1 x y)))
  '(((x.0 x1) (y.0 y1))
    ((x.0 x2) (y.0 y2))))

(test-check "Rinf+R1"
  (solve 7 (x y)
    (any-union (Rinf x y) (R1 x y)))
  '(((x.0 z) (y.0 z))
    ((x.0 x1) (y.0 y1))
    ((x.0 (s z)) (y.0 (s z)))
    ((x.0 x2) (y.0 y2))
    ((x.0 (s (s z))) (y.0 (s (s z))))
    ((x.0 (s (s (s z)))) (y.0 (s (s (s z)))))
    ((x.0 (s (s (s (s z))))) (y.0 (s (s (s (s z))))))))

(test-check "R1+RInf"
  (solve 7 (x y)
    (any-union (R1 x y) (Rinf x y)))
  '(((x.0 x1) (y.0 y1))
    ((x.0 z) (y.0 z))
    ((x.0 x2) (y.0 y2))
    ((x.0 (s z)) (y.0 (s z)))
    ((x.0 (s (s z))) (y.0 (s (s z))))
    ((x.0 (s (s (s z)))) (y.0 (s (s (s z)))))
    ((x.0 (s (s (s (s z))))) (y.0 (s (s (s (s z))))))))


; Infinitary relation Rinf2
; r(z,z).
; r(s(s(X)),s(s(Y))) :- r(X,Y).
; Rinf2 overlaps with Rinf in the infinite number of points

(define Rinf2
  (extend-relation (a1 a2)
    (fact () 'z 'z)
    (relation (x y t1 t2)
      (to-show t1 t2)
      (all
	(== t1 `(s (s ,x)))
	(== t2 `(s (s ,y)))
	(Rinf2 x y)))))

(test-check "Rinf2"
  (solve 5 (x y) (Rinf2 x y))
  '(((x.0 z) (y.0 z))
    ((x.0 (s (s z))) (y.0 (s (s z))))
    ((x.0 (s (s (s (s z))))) (y.0 (s (s (s (s z))))))
    ((x.0 (s (s (s (s (s (s z)))))))
     (y.0 (s (s (s (s (s (s z))))))))
    ((x.0 (s (s (s (s (s (s (s (s z)))))))))
     (y.0 (s (s (s (s (s (s (s (s z))))))))))))

(test-check "Rinf+Rinf2"
  (solve 9 (x y)
    (any-union (Rinf x y) (Rinf2 x y)))
  '(((x.0 z) (y.0 z))
    ((x.0 (s z)) (y.0 (s z)))
    ((x.0 (s (s z))) (y.0 (s (s z))))
    ((x.0 (s (s (s (s z))))) (y.0 (s (s (s (s z))))))
    ((x.0 (s (s (s z)))) (y.0 (s (s (s z)))))
    ((x.0 (s (s (s (s (s (s z)))))))
     (y.0 (s (s (s (s (s (s z))))))))
    ((x.0 (s (s (s (s (s (s (s (s z)))))))))
     (y.0 (s (s (s (s (s (s (s (s z))))))))))
    ((x.0 (s (s (s (s (s z)))))) (y.0 (s (s (s (s (s z)))))))
    ((x.0 (s (s (s (s (s (s (s (s (s (s z)))))))))))
     (y.0 (s (s (s (s (s (s (s (s (s (s z))))))))))))))

(test-check "Rinf2+Rinf"
  (solve 9 (x y)
    (any-union (Rinf2 x y) (Rinf x y)))
  '(((x.0 z) (y.0 z))
    ((x.0 (s z)) (y.0 (s z)))
    ((x.0 (s (s z))) (y.0 (s (s z))))
    ((x.0 (s (s (s z)))) (y.0 (s (s (s z)))))
    ((x.0 (s (s (s (s z))))) (y.0 (s (s (s (s z))))))
    ((x.0 (s (s (s (s (s z)))))) (y.0 (s (s (s (s (s z)))))))
    ((x.0 (s (s (s (s (s (s z)))))))
     (y.0 (s (s (s (s (s (s z))))))))
    ((x.0 (s (s (s (s (s (s (s z))))))))
     (y.0 (s (s (s (s (s (s (s z)))))))))
    ((x.0 (s (s (s (s (s (s (s (s z)))))))))
     (y.0 (s (s (s (s (s (s (s (s z))))))))))))

;equality predicate: X == Y in Prolog
;if X is a var, then X == Y holds only if Y
;is the same var
(define *equal?
  (lambda (x y)
    (cond
      [(and (var? x) (var? y)) (eq? x y)]
      [(var? x) #f]                     ; y is not a var
      [(var? y) #f]                     ; x is not a var
      [else (equal? x y)])))

; extend-relation-with-recur-limit LIMIT VARS RELS -> REL
; This is a variation of 'extend-relation' that makes sure
; that the extended relation is not recursively entered more
; than LIMIT times. The form extend-relation-with-recur-limit
; can be used to cut a left-recursive relation, and to implement
; an iterative deepening strategy.
; extend-relation-with-recur-limit must be a special form
; because we need to define the depth-counter-var
; outside of relations' lambda (so we count the recursive invocations
; for all arguments).
(define-syntax extend-relation-with-recur-limit
  (syntax-rules ()
    ((_ limit ids rel ...)
      (let ((depth-counter-var (logical-variable '*depth-counter*)))
	(lambda ids
	  (let ((ant (any (rel . ids) ...)))
	    (lambda@ (sk fk subst)
	      (cond
		[(assq depth-counter-var subst)
		  => (lambda (cmt)
		       (let ([counter (commitment->term cmt)])
			 (if (>= counter limit)
			   (fk)
			   (let ([s (extend-subst depth-counter-var
				      (+ counter 1) subst)])
			     (@ ant sk fk s)))))]
		[else
		  (let ([s (extend-subst depth-counter-var 0 subst)])
		    (@ ant sk fk s))]))))))
    ))

(cout nl "Append with limited depth" nl)
; In Prolog, we normally write:
; append([],L,L).
; append([X|L1],L2,[X|L3]) :- append(L1,L2,L3).
;
; If we switch the clauses, we get non-termination.
; In our system, it doesn't matter!

(define extend-clause-1
  (relation (l)
    (to-show '() l l)))

(define extend-clause-2
  (relation (x l1 l2 l3)
    (to-show `(,x . ,l1) l2 `(,x . ,l3))
    (extend-rel l1 l2 l3)))

(define extend-rel
  (extend-relation-with-recur-limit 5 (a b c)
    extend-clause-1
    extend-clause-2))

; Note (solve 100 ...)
; Here 100 is just a large number: we want to print all solutions
(cout nl "Extend: clause1 first: " 
  (solve 100 (a b c) (extend-rel a b c))
  nl)

(define extend-rel
  (extend-relation-with-recur-limit 3 (a b c)
    extend-clause-2
    extend-clause-1))

(cout nl "Extend: clause2 first. In Prolog, it would diverge!: " 
  (solve 100 (a b c) (extend-rel a b c)) nl)


;(exit 0)

(define base-+-as-relation
  (fact (n) 'zero n n))

(define recursive-+-as-relation
  (relation (n1 n2 n3)
    (to-show `(succ ,n1) n2 `(succ ,n3))
    (+-as-relation n1 n2 n3)))

(define +-as-relation
  (extend-relation (a1 a2 a3)
    base-+-as-relation
    recursive-+-as-relation))

(test-check "Addition"
  (solve 20 (x y)
    (+-as-relation x y '(succ (succ (succ (succ (succ zero)))))))
  '(((x.0 zero) (y.0 (succ (succ (succ (succ (succ zero)))))))
    ((x.0 (succ zero)) (y.0 (succ (succ (succ (succ zero))))))
    ((x.0 (succ (succ zero))) (y.0 (succ (succ (succ zero)))))
    ((x.0 (succ (succ (succ zero)))) (y.0 (succ (succ zero))))
    ((x.0 (succ (succ (succ (succ zero))))) (y.0 (succ zero)))
    ((x.0 (succ (succ (succ (succ (succ zero)))))) (y.0 zero))))

(newline)

;(exit 0)


