#! /usr/local/bin/gsi -:u
":"; exec /usr/local/bin/gsi -:u $0
;			The Binder Application
;
; Test cases:
; http://www.metnet.navy.mil/Metcast/Auth-use-cases.html
;
; This is Gambit code, because socket support on Bigloo is wanting.
; Flushing the output port does not flushes the packet,
; especially on Linux.
;
; $Id: binder-tests.scm,v 1.2 2004/07/10 05:28:45 oleg Exp $

;(include "myenv.scm")

; (module binder-tests
;   (include "myenv-bigloo.scm")
;   (include "srfi-13-local.scm")
;   (include "util.scm")
; )

(define soutei-host "tcp://localhost:1500")

; Submit the request to a remote SOUTEI application and return its result
(define (do-test req)
  (let ((connection (##open-input-output-file soutei-host)))
    (cerr nl "connection established" nl)
    (write req connection) (display "\r\n" connection)
    (flush-output connection)

    (let ((response (read connection)))
      (assert (pair? response) (boolean? (cadr response)))
      (cadr response))))

; Scenario "Local-MEMO" "Local access to MEMO"

; group "local-MEMO-1"

(assert
  (do-test 
    '(req-local-MEMO-1
       (Channel "MEMO")
       (Access write)
       (IP "10.10.1.1")
       )))

(assert
  (do-test 
    '(req-local-MEMO-1
       (Channel "MEMO")
       (Access read)
       (IP "10.10.1.1")
       )))

(assert
  (do-test 
    '(req-local-MEMO-1
       (Channel "MEMO")
       (Access write)
       (IP "192.168.1.1") ; Also internal
       )))

; Scenario "Remote-MEMO" "Remote access to MEMO"

; group "MEMO-Basic-authentication-checks"

(assert
  (not (do-test 
    '(req-MEMO-Basic-authentication-checks
       (Channel "MEMO")
       (Access read)
       (IP "204.1.2.3") ; Some external host
       ))))

(assert
  (not (do-test 
    '(req-MEMO-Basic-authentication-checks
       (Channel "MEMO")
       (Access read)
       (IP "204.1.2.3") ; Some external host
       (Subject "abcdef") ; Dean
       ))))


; group "MEMO-remote-OK"

(assert
  (do-test 
    '(req-MEMO-remote-OK
       (Channel "MEMO")
       (Access write)
       (IP "204.1.2.3") ; Some external host
       (Subject "0123456789") ; Joe
       )))

(assert
  (do-test 
    '(req-MEMO-remote-OK
       (Channel "MEMO")
       (Access read)
       (IP "204.1.2.3") ; Some external host
       (Subject "0123456789") ; Joe
       )))


; Scenario "DEMO-IMG-simple"
; Access to DEMO-IMG by its administrator. Simple authorizations."

; group "init-tests-DEMO-IMG"

(assert
  (not (do-test 
    '(req-init-tests-DEMO-IMG
       (Channel "DEMO-IMG")
       (Access write)
       (IP "10.10.1.1")
       ))))

(assert
  (not (do-test 
    '(req-init-tests-DEMO-IMG
       (Channel "DEMO-IMG")
       (Access read)
       (IP "204.1.2.3") ; Some external host
       ))))

(assert
  (not (do-test 
    '(req-init-tests-DEMO-IMG
       (Channel "DEMO-IMG")
       (Access read)
       (IP "204.1.2.3") ; Some external host
       (Subject "0123456789") ; Joe
       ))))

(assert
  (not (do-test 
    '(req-init-tests-DEMO-IMG
       (Channel "DEMO-IMG")
       (Access read)
       (IP "204.1.2.3") ; Some external host
       (Subject "abcdef") ; Dean
       ))))

; group "Dean-allows-himself"

(assert
  (do-test 
    '(req-Dean-allows-himself
       (Assertion-once
	 "abcdef" ;Dean
	 "may(?,?,?).")
       (Channel "DEMO-IMG")
       (Access write)
       (IP "172.10.10.1")
       (Subject "abcdef") ; Dean
     )))

(assert
  (do-test 
    '(req-Dean-allows-himself
       (Assertion-once
	 "abcdef" ;Dean
	 "may(?,?,?).")
       (Channel "DEMO-IMG")
       (Access read)
       (IP "172.10.10.1")
       (Subject "abcdef") ; Dean
     )))

; group "DEMO-IMG-check-rejections"

(assert
  (not (do-test 
    '(req-DEMO-IMG-check-rejections
       (Channel "DEMO-IMG")
       (Access read)
       (IP "204.1.2.3") ; Some external host
       (Subject "0123456789") ; Joe
       ))))

(assert
  (not (do-test 
    '(req-DEMO-IMG-check-rejections
       (Assertion-once
	 "abcdef" ;Dean
	 "may(?,?,?).")
       (Channel "MEMO")
       (Access read)
       (IP "172.10.10.1")
       (Subject "abcdef") ; Dean
     ))))

; 	(tc "Joe tries to use Dean's attribute certificate from " 
; 	  (group-ref "Dean-allows-himself")
; 	  " to read DEMO-IMG. User Joe sends the certificate in an
; HTTP header and or? TLS")

(assert
  (not (do-test 
    '(req-DEMO-IMG-check-rejections
       (Channel "DEMO-IMG")
       (Access read)
       (IP "10.10.1.1")
       (Subject "abcdef") ; Dean
     ))))

; group "DEMO-IMG-simple-authorization"
(assert
  (do-test 
    '(req-DEMO-IMG-simple-authorization
       (Assertion-once ; for Aaron
	 "abcdef" ;Dean
	 "may(?,?,?).")
       (Channel "DEMO-IMG")
       (Access read)
       (IP "172.10.10.1")
       (Subject "57899Zqvfy") ; Aaron
     )))


; Scenario "DEMO-IMG-delegation-1" "Access delegation via assertions."

; group "demo-img-delegation-1"
(assert
  (do-test 
    `(req-demo-img-delegation-1
       (Assertion-cache
	 "abcdef" ;Dean
	 ,(string-append
	   "may(channel,?,read) :- known_user(Eric)." nl
	   "known_user(Eric) :- 
               application says pubkey_fingerprint(\"dddddd\")." nl))
       ; validity period of one hour
       (Channel "DEMO-IMG")
       (Access read)
       (IP "172.10.10.1")
       (Subject "dddddd") ; Eric
     )))

(assert
  (do-test 
    '(req-demo-img-delegation-1
       ; Using the cached assertion
       (Channel "DEMO-IMG")
       (Access read)
       (IP "204.1.2.3") ; Some external host
       (Subject "dddddd") ; Eric
     )))

; group "demo-img-delegation-1-rej"
(assert
  (not (do-test 
    '(req-demo-img-delegation-1-rej
       ; Using the cached assertion
       (Channel "DEMO-IMG")
       (Access read)
       (IP "204.1.2.3") ; Some external host
       (Subject "0123456789") ; Joe
       ))))

(assert
  (not (do-test 
    `(req-demo-img-delegation-1-rej
       (Assertion-cache
	 "abcdef" ;Dean
	 ,(string-append
	   "may(channel,?,read) :- known_user(Eric)." nl
	   "known_user(Eric) :- 
               application says pubkey_fingerprint(\"dddddd\")." nl))
       (Channel "DEMO-IMG")
       (Access write)
       (IP "204.1.2.3") ; Some external host
       (Subject "dddddd") ; Eric
     ))))

(assert
  (not (do-test 
    `(req-demo-img-delegation-1-rej
       (Assertion-cache
	 "abcdef" ;Dean
	 ,(string-append
	   "may(channel,?,read) :- known_user(Eric)." nl
	   "known_user(Eric) :- 
               application says pubkey_fingerprint(\"dddddd\")." nl))
       (Channel "MEMO")
       (Access read)
       (IP "204.1.2.3") ; Some external host
       (Subject "dddddd") ; Eric
     ))))

; 	(tc "Eric attempts to read DEMO-IMG after one hour has passed"))

; group "demo-img-delegation-1-rej-2"

(assert
  (not (do-test 
    `(req-demo-img-delegation-1-rej-2
       (Assertion-cache
	 "abcdef" ;Dean
	 ,(string-append
	   "may(channel,?,read) :- known_user(Eric)." nl
	   "known_user(Eric) :- 
               application says pubkey_fingerprint(\"dddddd\")." nl))
       (Channel "DEMO-IMG")
       (Access read)
       (IP "204.1.2.3") ; Some external host
       (Subject "0123456789") ; Joe
     ))))


; Scenario "DEMO-IMG-delegation-2" "Chained delegation via assertions."

; group "demo-img-delegation-2"

(assert
  (do-test 
    '(req-demo-img-delegation-2
       (Assertion-cache
	 "abcdef" ;Dean
	 "may(?c,?n,read) :- \"eeeeee\" says may(?c,?n,?).")
       (Assertion-once ; To Greg
	 "eeeeee" ;Ryan
	 "may(?,?,?).")
       (Channel "DEMO-IMG")
       (Access read)
       (IP "204.1.2.3") ; Some external host
       (Subject "Ggsdjgsdjgsdj") ; Greg
     )))

(assert
  (do-test 
    '(req-demo-img-delegation-2
       ; Using the cached assertion
       (Assertion-once ; To Greg
	 "eeeeee" ;Ryan
	 "may(?,?,?).")
       (Channel "DEMO-IMG")
       (Access read)
       (IP "204.1.2.3") ; Some external host
       (Subject "Ggsdjgsdjgsdj") ; Greg
     )))

; group "demo-img-delegation-2-rej"

(assert
  (not (do-test 
    '(req-demo-img-delegation-2-rej
       (Channel "DEMO-IMG")
       (Access read)
       (IP "204.1.2.3") ; Some external host
       (Subject "0123456789") ; Joe
     ))))

(assert
  (not (do-test 
    '(req-demo-img-delegation-2-rej
       (Assertion-cache
	 "abcdef" ;Dean
	 "may(?c,?n,read) :- \"eeeeee\" says may(?c,?n,?).")
       (Assertion-once ; To Greg
	 "eeeeee" ;Ryan
	 "may(?,?,?).")
       (Channel "DEMO-IMG")
       (Access write)
       (IP "204.1.2.3") ; Some external host
       (Subject "Ggsdjgsdjgsdj") ; Greg
     ))))

(assert
  (not (do-test 
    '(req-demo-img-delegation-2-rej
       (Assertion-cache
	 "abcdef" ;Dean
	 "may(?c,?n,read) :- \"eeeeee\" says may(?c,?n,?).")
       (Assertion-once ; To Greg
	 "eeeeee" ;Ryan
	 "may(?,?,?).")
       (Channel "MEMO")
       (Access read)
       (IP "204.1.2.3") ; Some external host
       (Subject "Ggsdjgsdjgsdj") ; Greg
     ))))


(cerr nl "All tests passed" nl)


